"""Class definitions for multiprocessing Jobs"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Collection, Dict, List, NamedTuple, Set, Tuple, Union

from ..corpus.classes import File, Speaker, Utterance
from ..helper import output_mapping, save_scp

if TYPE_CHECKING:
    from ..aligner.adapting import AdaptingAligner
    from ..aligner.base import BaseAligner
    from ..config import FeatureConfig
    from ..config.align_config import AlignConfig, ConfigDict
    from ..corpus import Corpus
    from ..dictionary import (
        Dictionary,
        DictionaryData,
        MappingType,
        ReversedMappingType,
        WordsType,
    )
    from ..segmenter import Segmenter
    from ..speaker_classifier import SpeakerClassifier
    from ..trainers import BaseTrainer, IvectorExtractorTrainer, LdaTrainer, SatTrainer
    from ..transcriber import Transcriber
    from ..validator import CorpusValidator

    ConfigType = Union[BaseTrainer, AlignConfig]
    FmllrConfigType = Union[SatTrainer, AlignConfig]
    LdaConfigType = Union[LdaTrainer, AlignConfig]

    IterationType = Union[str, int]

    AlignerType = Union[BaseTrainer, BaseAligner]

__all__ = [
    "Job",
    "AlignArguments",
    "VadArguments",
    "SegmentVadArguments",
    "CreateHclgArguments",
    "AccGlobalStatsArguments",
    "AccStatsArguments",
    "AccIvectorStatsArguments",
    "AccStatsTwoFeatsArguments",
    "AliToCtmArguments",
    "MfccArguments",
    "ScoreArguments",
    "DecodeArguments",
    "PhoneCtmArguments",
    "CombineCtmArguments",
    "CleanupWordCtmArguments",
    "NoCleanupWordCtmArguments",
    "LmRescoreArguments",
    "AlignmentImprovementArguments",
    "ConvertAlignmentsArguments",
    "CalcFmllrArguments",
    "CalcLdaMlltArguments",
    "GmmGselectArguments",
    "FinalFmllrArguments",
    "LatGenFmllrArguments",
    "FmllrRescoreArguments",
    "TreeStatsArguments",
    "LdaAccStatsArguments",
    "MapAccStatsArguments",
    "GaussToPostArguments",
    "ClassifySpeakersArguments",
    "InitialFmllrArguments",
    "ExtractIvectorsArguments",
    "ExportTextGridArguments",
    "CompileTrainGraphsArguments",
    "CompileInformationArguments",
    "CompileUtteranceTrainGraphsArguments",
    "MonoAlignEqualArguments",
    "TestUtterancesArguments",
    "CarpaLmRescoreArguments",
    "GeneratePronunciationsArguments",
]


class VadArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.features.compute_vad_func`"""

    log_path: str
    dictionaries: List[str]
    feats_scp_paths: Dict[str, str]
    vad_scp_paths: Dict[str, str]
    vad_options: ConfigDict


class MfccArguments(NamedTuple):
    """
    Arguments for :func:`~montreal_forced_aligner.multiprocessing.features.mfcc_func`
    """

    log_path: str
    dictionaries: List[str]
    feats_scp_paths: Dict[str, str]
    lengths_paths: Dict[str, str]
    segment_paths: Dict[str, str]
    wav_paths: Dict[str, str]
    mfcc_options: ConfigDict


class CompileTrainGraphsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_train_graphs_func`"""

    log_path: str
    dictionaries: List[str]
    tree_path: str
    model_path: str
    text_int_paths: Dict[str, str]
    disambig_paths: Dict[str, str]
    lexicon_fst_paths: Dict[str, str]
    fst_scp_paths: Dict[str, str]


class MonoAlignEqualArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.mono_align_equal_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    fst_scp_paths: Dict[str, str]
    ali_ark_paths: Dict[str, str]
    acc_paths: Dict[str, str]
    model_path: str


class AccStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.acc_stats_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    acc_paths: Dict[str, str]
    model_path: str


class AlignArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.align_func`"""

    log_path: str
    dictionaries: List[str]
    fst_scp_paths: Dict[str, str]
    feature_strings: Dict[str, str]
    model_path: str
    ali_paths: Dict[str, str]
    score_paths: Dict[str, str]
    loglike_paths: Dict[str, str]
    align_options: ConfigDict


class CompileInformationArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_information_func`"""

    align_log_paths: str


class AliToCtmArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.ali_to_ctm_func`"""

    log_path: str
    dictionaries: List[str]
    ali_paths: Dict[str, str]
    text_int_paths: Dict[str, str]
    word_boundary_int_paths: Dict[str, str]
    frame_shift: float
    model_path: str
    ctm_paths: Dict[str, str]
    word_mode: bool


class CleanupWordCtmArguments(NamedTuple):
    """Arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.CleanupWordCtmProcessWorker`"""

    ctm_paths: Dict[str, str]
    dictionaries: List[str]
    utterances: Dict[str, Dict[str, Utterance]]
    dictionary_data: Dict[str, DictionaryData]


class NoCleanupWordCtmArguments(NamedTuple):
    """Arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.NoCleanupWordCtmProcessWorker`"""

    ctm_paths: Dict[str, str]
    dictionaries: List[str]
    utterances: Dict[str, Dict[str, Utterance]]
    dictionary_data: Dict[str, DictionaryData]


class PhoneCtmArguments(NamedTuple):
    """Arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.PhoneCtmProcessWorker`"""

    ctm_paths: Dict[str, str]
    dictionaries: List[str]
    utterances: Dict[str, Dict[str, Utterance]]
    reversed_phone_mappings: Dict[str, ReversedMappingType]
    positions: Dict[str, List[str]]


class CombineCtmArguments(NamedTuple):
    """Arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.CombineProcessWorker`"""

    dictionaries: List[str]
    files: Dict[str, File]
    dictionary_data: Dict[str, DictionaryData]
    cleanup_textgrids: bool


class ExportTextGridArguments(NamedTuple):
    """Arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.ExportTextGridProcessWorker`"""

    files: Dict[str, File]
    frame_shift: int
    output_directory: str
    backup_output_directory: str


class TreeStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.tree_stats_func`"""

    log_path: str
    dictionaries: List[str]
    ci_phones: str
    model_path: str
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    treeacc_paths: Dict[str, str]


class ConvertAlignmentsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.convert_alignments_func`"""

    log_path: str
    dictionaries: List[str]
    model_path: str
    tree_path: str
    align_model_path: str
    ali_paths: Dict[str, str]
    new_ali_paths: Dict[str, str]


class AlignmentImprovementArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compute_alignment_improvement_func`"""

    log_path: str
    dictionaries: List[str]
    model_path: str
    text_int_paths: Dict[str, str]
    word_boundary_paths: Dict[str, str]
    ali_paths: Dict[str, str]
    frame_shift: int
    reversed_phone_mappings: Dict[str, Dict[int, str]]
    positions: Dict[str, List[str]]
    phone_ctm_paths: Dict[str, str]


class CalcFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.calc_fmllr_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    ali_model_path: str
    model_path: str
    spk2utt_paths: Dict[str, str]
    trans_paths: Dict[str, str]
    fmllr_options: ConfigDict


class AccStatsTwoFeatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.acc_stats_two_feats_func`"""

    log_path: str
    dictionaries: List[str]
    ali_paths: Dict[str, str]
    acc_paths: Dict[str, str]
    model_path: str
    feature_strings: Dict[str, str]
    si_feature_strings: Dict[str, str]


class LdaAccStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.lda_acc_stats_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    lda_options: ConfigDict
    acc_paths: Dict[str, str]


class CalcLdaMlltArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.calc_lda_mllt_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    lda_options: ConfigDict
    macc_paths: Dict[str, str]


class MapAccStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.map_acc_stats_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    model_path: str
    ali_paths: Dict[str, str]
    acc_paths: Dict[str, str]


class GmmGselectArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.gmm_gselect_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ivector_options: ConfigDict
    dubm_model: str
    gselect_paths: Dict[str, str]


class AccGlobalStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.acc_global_stats_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ivector_options: ConfigDict
    gselect_paths: Dict[str, str]
    acc_paths: Dict[str, str]
    dubm_path: str


class GaussToPostArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.gauss_to_post_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ivector_options: ConfigDict
    post_paths: Dict[str, str]
    dubm_path: str


class AccIvectorStatsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.acc_ivector_stats_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ivector_options: ConfigDict
    ie_path: str
    post_paths: Dict[str, str]
    acc_init_paths: Dict[str, str]


class ExtractIvectorsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.extract_ivectors_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ivector_options: ConfigDict
    ali_paths: Dict[str, str]
    ie_path: str
    ivector_paths: Dict[str, str]
    weight_paths: Dict[str, str]
    model_path: str
    dubm_path: str


class CompileUtteranceTrainGraphsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_utterance_train_graphs_func`"""

    log_path: str
    dictionaries: List[str]
    disambig_int_paths: Dict[str, str]
    disambig_L_fst_paths: Dict[str, str]
    fst_paths: Dict[str, str]
    graphs_paths: Dict[str, str]
    model_path: str
    tree_path: str


class TestUtterancesArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.test_utterances_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    words_paths: Dict[str, str]
    graphs_paths: Dict[str, str]
    text_int_paths: Dict[str, str]
    edits_paths: Dict[str, str]
    out_int_paths: Dict[str, str]
    model_path: str


class SegmentVadArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.segment_vad_func`"""

    dictionaries: List[str]
    vad_paths: Dict[str, str]
    segmentation_options: ConfigDict


class ClassifySpeakersArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.classify_speakers_func`"""

    log_path: str
    dictionaries: List[str]
    model_path: str
    labels_path: str
    ivector_paths: Dict[str, str]


class GeneratePronunciationsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.pronunciations.generate_pronunciations_func`"""

    log_path: str
    dictionaries: List[str]
    text_int_paths: Dict[str, str]
    word_boundary_paths: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    pron_paths: Dict[str, str]


class CreateHclgArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.create_hclg_func`"""

    log_path: str
    working_directory: str
    path_template: str
    words_path: str
    carpa_path: str
    small_arpa_path: str
    medium_arpa_path: str
    big_arpa_path: str
    model_path: str
    disambig_L_path: str
    disambig_int_path: str
    hclg_options: ConfigDict
    words_mapping: MappingType

    @property
    def hclg_path(self) -> str:
        return self.path_template.format(file_name="HCLG")


class DecodeArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.decode_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    decode_options: ConfigDict
    model_path: str
    lat_paths: Dict[str, str]
    words_paths: Dict[str, str]
    hclg_paths: Dict[str, str]


class ScoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.score_func`"""

    log_path: str
    dictionaries: List[str]
    score_options: ConfigDict
    lat_paths: Dict[str, str]
    rescored_lat_paths: Dict[str, str]
    carpa_rescored_lat_paths: Dict[str, str]
    words_paths: Dict[str, str]
    tra_paths: Dict[str, str]


class LmRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.lm_rescore_func`"""

    log_path: str
    dictionaries: List[str]
    lm_rescore_options: ConfigDict
    lat_paths: Dict[str, str]
    rescored_lat_paths: Dict[str, str]
    old_g_paths: Dict[str, str]
    new_g_paths: Dict[str, str]


class CarpaLmRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.carpa_lm_rescore_func`"""

    log_path: str
    dictionaries: List[str]
    lat_paths: Dict[str, str]
    rescored_lat_paths: Dict[str, str]
    old_g_paths: Dict[str, str]
    new_g_paths: Dict[str, str]


class InitialFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.initial_fmllr_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    model_path: str
    fmllr_options: ConfigDict
    pre_trans_paths: Dict[str, str]
    lat_paths: Dict[str, str]
    spk2utt_paths: Dict[str, str]


class LatGenFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.lat_gen_fmllr_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    model_path: str
    decode_options: ConfigDict
    words_paths: Dict[str, str]
    hclg_paths: Dict[str, str]
    tmp_lat_paths: Dict[str, str]


class FinalFmllrArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.final_fmllr_est_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    model_path: str
    fmllr_options: ConfigDict
    trans_paths: Dict[str, str]
    spk2utt_paths: Dict[str, str]
    tmp_lat_paths: Dict[str, str]


class FmllrRescoreArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.fmllr_rescore_func`"""

    log_path: str
    dictionaries: List[str]
    feature_strings: Dict[str, str]
    model_path: str
    fmllr_options: ConfigDict
    tmp_lat_paths: Dict[str, str]
    final_lat_paths: Dict[str, str]


class Job:
    """
    Class representing information about corpus jobs that will be run in parallel.
    Jobs have a set of speakers that they will process, along with all files and utterances associated with that speaker.
    As such, Jobs also have a set of dictionaries that the speakers use, and argument outputs are largely dependent on
    the pronunciation dictionaries in use.

    Parameters
    ----------
    name: int
        Job number is the job's identifier

    Attributes
    ----------
    speakers: List[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        List of speakers associated with this job
    dictionaries: Set[:class:`~montreal_forced_aligner.dictionary.Dictionary`]
        Set of dictionaries that the job's speakers use
    subset_utts: Set[:class:`~montreal_forced_aligner.corpus.classes.Utterance`]
        When trainers are just using a subset of the corpus, the subset of utterances on each job will be set and used to
        filter the job's utterances
    subset_speakers: Set[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        When subset_utts is set, this property will be calculated as the subset of speakers that the utterances correspond to
    subset_dictionaries: Set[:class:`~montreal_forced_aligner.dictionary.Dictionary`]
        Subset of dictionaries that the subset of speakers use

    """

    def __init__(self, name: int):
        self.name = name
        self.speakers: List[Speaker] = []
        self.dictionaries: Set[Dictionary] = set()

        self.subset_utts: Set[Utterance] = set()
        self.subset_speakers: Set[Speaker] = set()
        self.subset_dictionaries: Set[Dictionary] = set()

    def add_speaker(self, speaker: Speaker) -> None:
        """
        Add a speaker to a job

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            Speaker to add
        """
        self.speakers.append(speaker)
        self.dictionaries.add(speaker.dictionary)

    def set_subset(self, subset_utts: Collection[Utterance]) -> None:
        """
        Set the current subset for the trainer

        Parameters
        ----------
        subset_utts: List[:class:`~montreal_forced_aligner.corpus.classes.Utterance`]
            Subset of utterances for this job to use
        """
        if subset_utts is None:
            self.subset_utts = set()
            self.subset_speakers = set()
            self.subset_dictionaries = set()
        else:
            self.subset_utts = set(subset_utts)
            self.subset_speakers = {u.speaker for u in subset_utts if u.speaker in self.speakers}
            self.subset_dictionaries = {s.dictionary for s in self.subset_speakers}

    def text_scp_data(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Generate the job's data for Kaldi's text scp files

        Returns
        -------
        Dict[str, Dict[str, List[str]]]
            Text for each utterance, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if u.ignored:
                    continue
                if self.subset_utts and u not in self.subset_utts:
                    continue
                if not u.text:
                    continue
                data[key][u.name] = u.text_for_scp()
        return data

    def text_int_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's text int scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Text converted to integer IDs for each utterance, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                continue
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if self.subset_utts and u not in self.subset_utts:
                    continue
                if u.ignored:
                    continue
                if not u.text:
                    continue
                data[key][u.name] = " ".join(map(str, u.text_int_for_scp()))
        return data

    def wav_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's wav scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Wav scp strings for each file, per dictionary name
        """
        data = {}
        done = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
                done[key] = set()
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if u.ignored:
                    continue
                if self.subset_utts and u not in self.subset_utts:
                    continue
                if not u.is_segment:
                    data[key][u.name] = u.file.for_wav_scp()
                elif u.file.name not in done:
                    data[key][u.file.name] = u.file.for_wav_scp()
                    done[key].add(u.file.name)
        return data

    def utt2spk_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's utt2spk scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Utterance to speaker mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if u.ignored:
                    continue
                if self.subset_utts and u not in self.subset_utts:
                    continue
                data[key][u.name] = s.name
        return data

    def feat_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's feature scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Utterance to feature archive ID mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if u.ignored:
                    continue
                if self.subset_utts and u not in self.subset_utts:
                    continue
                if u.features:
                    data[key][u.name] = u.features
        return data

    def spk2utt_scp_data(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Generate the job's data for Kaldi's spk2utt scp files

        Returns
        -------
        Dict[str, Dict[str, List[str]]]
            Speaker to utterance mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            data[key][s.name] = sorted(
                [
                    u.name
                    for u in s.utterances.values()
                    if not u.ignored and not (self.subset_utts and u not in self.subset_utts)
                ]
            )
        return data

    def cmvn_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's CMVN scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Speaker to CMVN mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            if s.cmvn:
                data[key][s.name] = s.cmvn
        return data

    def segments_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's segments scp files

        Returns
        -------
        Dict[str, Dict[str, str]]
            Utterance to segment mapping, per dictionary name
        """
        data = {}
        for s in self.speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary.name
            if key not in data:
                data[key] = {}
            if self.subset_speakers and s not in self.subset_speakers:
                continue
            for u in s.utterances.values():
                if u.ignored:
                    continue
                if self.subset_utts and u not in self.subset_utts:
                    continue
                if not u.is_segment:
                    continue
                data[key][u.name] = u.segment_for_scp()
        return data

    def construct_path_dictionary(
        self, directory: str, identifier: str, extension: str
    ) -> Dict[str, str]:
        """
        Helper function for constructing dictionary-dependent paths for the Job

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        Dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(
                directory, f"{identifier}.{dict_name}.{self.name}.{extension}"
            )
        return output

    def construct_dictionary_dependent_paths(
        self, directory: str, identifier: str, extension: str
    ) -> Dict[str, str]:
        """
        Helper function for constructing paths that depend only on the dictionaries of the job, and not the job name itself.
        These paths should be merged with all other jobs to get a full set of dictionary paths.

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        Dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(directory, f"{identifier}.{dict_name}.{extension}")
        return output

    @property
    def dictionary_count(self):
        """Number of dictionaries currently used"""
        if self.subset_dictionaries:
            return len(self.subset_dictionaries)
        return len(self.dictionaries)

    @property
    def current_dictionaries(self):
        """Current dictionaries depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return self.subset_dictionaries
        return self.dictionaries

    @property
    def current_dictionary_names(self):
        """Current dictionary names depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return sorted(x.name for x in self.subset_dictionaries)
        if self.dictionaries == {None}:
            return [None]
        return sorted(x.name for x in self.dictionaries)

    def set_feature_config(self, feature_config: FeatureConfig) -> None:
        """
        Set the feature configuration to use for the Job

        Parameters
        ----------
        feature_config: :class:`~montreal_forced_aligner.config.features.FeatureConfig`
            Feature configuration
        """
        self.feature_config = feature_config

    def construct_base_feature_string(self, corpus: Corpus, all_feats: bool = False) -> str:
        """
        Construct the base feature string independent of job name

        Used in initialization of MonophoneTrainer (to get dimension size) and IvectorTrainer (uses all feats)

        Parameters
        ----------
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use as the source
        all_feats: bool
            Flag for whether all features across all jobs should be taken into account

        Returns
        -------
        str
            Feature string
        """
        if all_feats:
            feat_path = os.path.join(corpus.output_directory, "feats.scp")
            utt2spk_path = os.path.join(corpus.output_directory, "utt2spk.scp")
            cmvn_path = os.path.join(corpus.output_directory, "cmvn.scp")
            feats = f"ark,s,cs:apply-cmvn --utt2spk=ark:{utt2spk_path} scp:{cmvn_path} scp:{feat_path} ark:- |"
            feats += " add-deltas ark:- ark:- |"
            return feats
        utt2spks = self.construct_path_dictionary(corpus.split_directory, "utt2spk", "scp")
        cmvns = self.construct_path_dictionary(corpus.split_directory, "cmvn", "scp")
        features = self.construct_path_dictionary(corpus.split_directory, "feats", "scp")
        for dict_name in self.current_dictionary_names:
            feat_path = features[dict_name]
            cmvn_path = cmvns[dict_name]
            utt2spk_path = utt2spks[dict_name]
            feats = f"ark,s,cs:apply-cmvn --utt2spk=ark:{utt2spk_path} scp:{cmvn_path} scp:{feat_path} ark:- |"
            if self.feature_config.deltas:
                feats += " add-deltas ark:- ark:- |"

            return feats

    def construct_feature_proc_strings(
        self,
        aligner: Union[AlignerType, SpeakerClassifier, Transcriber],
        speaker_independent: bool = False,
    ) -> Dict[str, str]:
        """
        Constructs a feature processing string to supply to Kaldi binaries, taking into account corpus features and the
        current working directory of the aligner (whether fMLLR or LDA transforms should be used, etc).

        Parameters
        ----------
        aligner: Union[AlignerType, :class:`~montreal_forced_aligner.speaker_classifier.SpeakerClassifier`, :class:`~montreal_forced_aligner.transcriber.Transcriber`]
            Aligner, Transcriber or other main utility class that uses the features
        speaker_independent: bool
            Flag for whether features should be speaker-independent regardless of the presence of fMLLR transforms

        Returns
        -------
        Dict[str, str]
            Feature strings per dictionary name
        """
        lda_mat_path = None
        fmllrs = {}
        if aligner.working_directory is not None:
            lda_mat_path = os.path.join(aligner.working_directory, "lda.mat")
            if not os.path.exists(lda_mat_path):
                lda_mat_path = None

            fmllrs = self.construct_path_dictionary(aligner.working_directory, "trans", "ark")
        utt2spks = self.construct_path_dictionary(aligner.data_directory, "utt2spk", "scp")
        cmvns = self.construct_path_dictionary(aligner.data_directory, "cmvn", "scp")
        features = self.construct_path_dictionary(aligner.data_directory, "feats", "scp")
        vads = self.construct_path_dictionary(aligner.data_directory, "vad", "scp")
        feat_strings = {}
        for dict_name in self.current_dictionary_names:
            feat_path = features[dict_name]
            cmvn_path = cmvns[dict_name]
            utt2spk_path = utt2spks[dict_name]
            fmllr_trans_path = None
            try:
                fmllr_trans_path = fmllrs[dict_name]
                if not os.path.exists(fmllr_trans_path):
                    fmllr_trans_path = None
            except KeyError:
                pass
            vad_path = vads[dict_name]
            if aligner.uses_voiced:
                feats = f"ark,s,cs:add-deltas scp:{feat_path} ark:- |"
                if aligner.uses_cmvn:
                    feats += " apply-cmvn-sliding --norm-vars=false --center=true --cmn-window=300 ark:- ark:- |"
                feats += f" select-voiced-frames ark:- scp,s,cs:{vad_path} ark:- |"
            elif not os.path.exists(cmvn_path) and aligner.uses_cmvn:
                feats = f"ark,s,cs:add-deltas scp:{feat_path} ark:- |"
                if aligner.uses_cmvn:
                    feats += " apply-cmvn-sliding --norm-vars=false --center=true --cmn-window=300 ark:- ark:- |"
            else:
                feats = f"ark,s,cs:apply-cmvn --utt2spk=ark:{utt2spk_path} scp:{cmvn_path} scp:{feat_path} ark:- |"
                if lda_mat_path is not None:
                    if not os.path.exists(lda_mat_path):
                        raise Exception(f"Could not find {lda_mat_path}")
                    feats += f" splice-feats --left-context={self.feature_config.splice_left_context} --right-context={self.feature_config.splice_right_context} ark:- ark:- |"
                    feats += f" transform-feats {lda_mat_path} ark:- ark:- |"
                elif aligner.uses_splices:
                    feats += f" splice-feats --left-context={self.feature_config.splice_left_context} --right-context={self.feature_config.splice_right_context} ark:- ark:- |"
                elif self.feature_config.deltas:
                    feats += " add-deltas ark:- ark:- |"

                if fmllr_trans_path is not None and not (
                    aligner.speaker_independent or speaker_independent
                ):
                    if not os.path.exists(fmllr_trans_path):
                        raise Exception(f"Could not find {fmllr_trans_path}")
                    feats += f" transform-feats --utt2spk=ark:{utt2spk_path} ark:{fmllr_trans_path} ark:- ark:- |"
            feat_strings[dict_name] = feats
        return feat_strings

    def compile_utterance_train_graphs_arguments(
        self, validator: CorpusValidator
    ) -> CompileUtteranceTrainGraphsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_utterance_train_graphs_func`

        Parameters
        ----------
        validator: :class:`~montreal_forced_aligner.validator.CorpusValidator`
            Validator

        Returns
        -------
        CompileUtteranceTrainGraphsArguments
            Arguments for processing
        """
        dictionary_paths = validator.dictionary.output_paths
        disambig_paths = {
            k: os.path.join(v, "phones", "disambiguation_symbols.int")
            for k, v in dictionary_paths.items()
        }
        lexicon_fst_paths = {
            k: os.path.join(v, "L_disambig.fst") for k, v in dictionary_paths.items()
        }
        return CompileUtteranceTrainGraphsArguments(
            os.path.join(
                validator.trainer.working_log_directory, f"utterance_fst.{self.name}.log"
            ),
            self.current_dictionary_names,
            disambig_paths,
            lexicon_fst_paths,
            self.construct_path_dictionary(validator.trainer.data_directory, "utt2fst", "scp"),
            self.construct_path_dictionary(
                validator.trainer.working_directory, "utterance_graphs", "fst"
            ),
            validator.trainer.current_model_path,
            validator.trainer.tree_path,
        )

    def test_utterances_arguments(self, validator: CorpusValidator) -> TestUtterancesArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.test_utterances_func`

        Parameters
        ----------
        validator: :class:`~montreal_forced_aligner.validator.CorpusValidator`
            Validator

        Returns
        -------
        TestUtterancesArguments
            Arguments for processing
        """
        dictionary_paths = validator.dictionary.output_paths
        words_paths = {k: os.path.join(v, "words.txt") for k, v in dictionary_paths.items()}
        return TestUtterancesArguments(
            os.path.join(validator.trainer.working_directory, f"utterance_fst.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(validator.trainer),
            words_paths,
            self.construct_path_dictionary(
                validator.trainer.working_directory, "utterance_graphs", "fst"
            ),
            self.construct_path_dictionary(validator.trainer.data_directory, "text", "int.scp"),
            self.construct_path_dictionary(validator.trainer.working_directory, "edits", "scp"),
            self.construct_path_dictionary(validator.trainer.working_directory, "aligned", "int"),
            validator.trainer.current_model_path,
        )

    def extract_ivector_arguments(
        self, ivector_extractor: SpeakerClassifier
    ) -> ExtractIvectorsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.extract_ivectors_func`

        Parameters
        ----------
        ivector_extractor: :class:`~montreal_forced_aligner.speaker_classifier.SpeakerClassifier`
            Speaker classifier

        Returns
        -------
        ExtractIvectorsArguments
            Arguments for processing
        """
        return ExtractIvectorsArguments(
            os.path.join(
                ivector_extractor.working_log_directory, f"extract_ivectors.{self.name}.log"
            ),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(ivector_extractor),
            ivector_extractor.ivector_options,
            self.construct_path_dictionary(ivector_extractor.working_directory, "ali", "ark"),
            ivector_extractor.ie_path,
            self.construct_path_dictionary(ivector_extractor.working_directory, "ivectors", "scp"),
            self.construct_path_dictionary(ivector_extractor.working_directory, "weights", "ark"),
            ivector_extractor.model_path,
            ivector_extractor.dubm_path,
        )

    def create_hclgs_arguments(self, transcriber: Transcriber) -> Dict[str, CreateHclgArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.create_hclg_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        Dict[str, CreateHclgArguments]
            Per dictionary arguments for HCLG
        """
        args = {}

        for dictionary in self.current_dictionaries:
            dict_name = dictionary.name
            args[dict_name] = CreateHclgArguments(
                os.path.join(transcriber.model_directory, "log", f"hclg.{dict_name}.log"),
                transcriber.model_directory,
                os.path.join(transcriber.model_directory, "{file_name}" + f".{dict_name}.fst"),
                os.path.join(transcriber.model_directory, f"words.{dict_name}.txt"),
                os.path.join(transcriber.model_directory, f"G.{dict_name}.carpa"),
                transcriber.language_model.small_arpa_path,
                transcriber.language_model.medium_arpa_path,
                transcriber.language_model.carpa_path,
                transcriber.model_path,
                dictionary.disambig_path,
                os.path.join(dictionary.phones_dir, "disambiguation_symbols.int"),
                transcriber.hclg_options,
                dictionary.words_mapping,
            )
        return args

    def decode_arguments(self, transcriber: Transcriber) -> DecodeArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.decode_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        DecodeArguments
            Arguments for processing
        """
        return DecodeArguments(
            os.path.join(transcriber.working_log_directory, f"decode.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(transcriber),
            transcriber.transcribe_config.decode_options,
            transcriber.alignment_model_path,
            self.construct_path_dictionary(transcriber.working_directory, "lat", "ark"),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "words", "txt"),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "HCLG", "fst"),
        )

    def score_arguments(self, transcriber: Transcriber) -> ScoreArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.score_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        ScoreArguments
            Arguments for processing
        """
        return ScoreArguments(
            os.path.join(transcriber.working_log_directory, f"score.{self.name}.log"),
            self.current_dictionary_names,
            transcriber.transcribe_config.score_options,
            self.construct_path_dictionary(transcriber.working_directory, "lat", "ark"),
            self.construct_path_dictionary(transcriber.working_directory, "lat.rescored", "ark"),
            self.construct_path_dictionary(
                transcriber.working_directory, "lat.carpa.rescored", "ark"
            ),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "words", "txt"),
            self.construct_path_dictionary(transcriber.evaluation_directory, "tra", "scp"),
        )

    def lm_rescore_arguments(self, transcriber: Transcriber) -> LmRescoreArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.lm_rescore_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        LmRescoreArguments
            Arguments for processing
        """
        return LmRescoreArguments(
            os.path.join(transcriber.working_log_directory, f"lm_rescore.{self.name}.log"),
            self.current_dictionary_names,
            transcriber.transcribe_config.lm_rescore_options,
            self.construct_path_dictionary(transcriber.working_directory, "lat", "ark"),
            self.construct_path_dictionary(transcriber.working_directory, "lat.rescored", "ark"),
            self.construct_dictionary_dependent_paths(
                transcriber.model_directory, "G.small", "fst"
            ),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "G.med", "fst"),
        )

    def carpa_lm_rescore_arguments(self, transcriber: Transcriber) -> CarpaLmRescoreArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.carpa_lm_rescore_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        CarpaLmRescoreArguments
            Arguments for processing
        """
        return CarpaLmRescoreArguments(
            os.path.join(transcriber.working_log_directory, f"carpa_lm_rescore.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(transcriber.working_directory, "lat.rescored", "ark"),
            self.construct_path_dictionary(
                transcriber.working_directory, "lat.carpa.rescored", "ark"
            ),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "G.med", "fst"),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "G", "carpa"),
        )

    def initial_fmllr_arguments(self, transcriber: Transcriber) -> InitialFmllrArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.initial_fmllr_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        InitialFmllrArguments
            Arguments for processing
        """
        return InitialFmllrArguments(
            os.path.join(transcriber.working_log_directory, f"initial_fmllr.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(transcriber),
            transcriber.model_path,
            transcriber.fmllr_options,
            self.construct_path_dictionary(transcriber.working_directory, "trans", "ark"),
            self.construct_path_dictionary(transcriber.working_directory, "lat", "ark"),
            self.construct_path_dictionary(transcriber.data_directory, "spk2utt", "scp"),
        )

    def lat_gen_fmllr_arguments(self, transcriber: Transcriber) -> LatGenFmllrArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.lat_gen_fmllr_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        LatGenFmllrArguments
            Arguments for processing
        """
        return LatGenFmllrArguments(
            os.path.join(transcriber.working_log_directory, f"lat_gen_fmllr.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(transcriber),
            transcriber.model_path,
            transcriber.transcribe_config.decode_options,
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "words", "txt"),
            self.construct_dictionary_dependent_paths(transcriber.model_directory, "HCLG", "fst"),
            self.construct_path_dictionary(transcriber.working_directory, "lat.tmp", "ark"),
        )

    def final_fmllr_arguments(self, transcriber: Transcriber) -> FinalFmllrArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.final_fmllr_est_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        FinalFmllrArguments
            Arguments for processing
        """
        return FinalFmllrArguments(
            os.path.join(transcriber.working_log_directory, f"final_fmllr.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(transcriber),
            transcriber.model_path,
            transcriber.fmllr_options,
            self.construct_path_dictionary(transcriber.working_directory, "trans", "ark"),
            self.construct_path_dictionary(transcriber.data_directory, "spk2utt", "scp"),
            self.construct_path_dictionary(transcriber.working_directory, "lat.tmp", "ark"),
        )

    def fmllr_rescore_arguments(self, transcriber: Transcriber) -> FmllrRescoreArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.transcription.fmllr_rescore_func`

        Parameters
        ----------
        transcriber: :class:`~montreal_forced_aligner.transcriber.Transcriber`
            Transcriber

        Returns
        -------
        FmllrRescoreArguments
            Arguments for processing
        """
        return FmllrRescoreArguments(
            os.path.join(transcriber.working_log_directory, f"fmllr_rescore.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(transcriber),
            transcriber.model_path,
            transcriber.fmllr_options,
            self.construct_path_dictionary(transcriber.working_directory, "lat.tmp", "ark"),
            self.construct_path_dictionary(transcriber.working_directory, "lat", "ark"),
        )

    def vad_arguments(self, corpus: Corpus) -> VadArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.features.compute_vad_func`

        Parameters
        ----------
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus

        Returns
        -------
        VadArguments
            Arguments for processing
        """
        return VadArguments(
            os.path.join(corpus.split_directory, "log", f"compute_vad.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(corpus.split_directory, "feats", "scp"),
            self.construct_path_dictionary(corpus.split_directory, "vad", "scp"),
            corpus.vad_config,
        )

    def segments_vad_arguments(self, segmenter: Segmenter) -> SegmentVadArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.segment_vad_func`

        Parameters
        ----------
        segmenter: :class:`~montreal_forced_aligner.segmenter.Segmenter`
            Segmenter

        Returns
        -------
        SegmentVadArguments
            Arguments for processing
        """
        return SegmentVadArguments(
            self.current_dictionary_names,
            self.construct_path_dictionary(segmenter.corpus.split_directory, "vad", "scp"),
            segmenter.segmentation_config.segmentation_options,
        )

    def mfcc_arguments(self, corpus: Corpus) -> MfccArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.features.mfcc_func`

        Parameters
        ----------
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus

        Returns
        -------
        MfccArguments
            Arguments for processing
        """
        return MfccArguments(
            os.path.join(corpus.split_directory, "log", f"make_mfcc.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(corpus.split_directory, "feats", "scp"),
            self.construct_path_dictionary(corpus.split_directory, "utterance_lengths", "scp"),
            self.construct_path_dictionary(corpus.split_directory, "segments", "scp"),
            self.construct_path_dictionary(corpus.split_directory, "wav", "scp"),
            self.feature_config.mfcc_options,
        )

    def acc_stats_arguments(self, aligner: AlignerType) -> AccStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.acc_stats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AccStatsArguments
            Arguments for processing
        """
        return AccStatsArguments(
            os.path.join(
                aligner.working_directory, "log", f"acc.{aligner.iteration}.{self.name}.log"
            ),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(
                aligner.working_directory, str(aligner.iteration), "acc"
            ),
            aligner.current_model_path,
        )

    def mono_align_equal_arguments(self, aligner: AlignerType) -> MonoAlignEqualArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.mono_align_equal_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        MonoAlignEqualArguments
            Arguments for processing
        """
        return MonoAlignEqualArguments(
            os.path.join(aligner.working_log_directory, f"mono_align_equal.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(aligner.working_directory, "fsts", "scp"),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.working_directory, "0", "acc"),
            aligner.current_model_path,
        )

    def align_arguments(self, aligner: AlignerType) -> AlignArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.align_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AlignArguments
            Arguments for processing
        """
        if aligner.iteration is not None:
            log_path = os.path.join(
                aligner.working_log_directory, f"align.{aligner.iteration}.{self.name}.log"
            )
        else:
            log_path = os.path.join(aligner.working_log_directory, f"align.{self.name}.log")
        return AlignArguments(
            log_path,
            self.current_dictionary_names,
            self.construct_path_dictionary(aligner.working_directory, "fsts", "scp"),
            self.construct_feature_proc_strings(aligner),
            aligner.alignment_model_path,
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.working_directory, "ali", "scores"),
            self.construct_path_dictionary(aligner.working_directory, "ali", "loglikes"),
            aligner.align_options,
        )

    def compile_information_arguments(self, aligner: AlignerType) -> CompileInformationArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_information_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        CompileInformationArguments
            Arguments for processing
        """
        if aligner.iteration is not None:
            log_path = os.path.join(
                aligner.working_log_directory, f"align.{aligner.iteration}.{self.name}.log"
            )
        else:
            log_path = os.path.join(aligner.working_log_directory, f"align.{self.name}.log")
        return CompileInformationArguments(log_path)

    def word_boundary_int_files(self) -> Dict[str, str]:
        """
        Generate mapping for dictionaries to word boundary int files

        Returns
        -------
        Dict[str, ReversedMappingType]
            Per dictionary word boundary int files
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = os.path.join(dictionary.phones_dir, "word_boundary.int")
        return data

    def reversed_phone_mappings(self) -> Dict[str, ReversedMappingType]:
        """
        Generate mapping for dictionaries to reversed phone mapping

        Returns
        -------
        Dict[str, ReversedMappingType]
            Per dictionary reversed phone mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.reversed_phone_mapping
        return data

    def reversed_word_mappings(self) -> Dict[str, ReversedMappingType]:
        """
        Generate mapping for dictionaries to reversed word mapping

        Returns
        -------
        Dict[str, ReversedMappingType]
            Per dictionary reversed word mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.reversed_word_mapping
        return data

    def words_mappings(self) -> Dict[str, MappingType]:
        """
        Generate mapping for dictionaries to word mapping

        Returns
        -------
        Dict[str, MappingType]
            Per dictionary word mapping
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.words_mapping
        return data

    def words(self) -> Dict[str, WordsType]:
        """
        Generate mapping for dictionaries to words

        Returns
        -------
        Dict[str, WordsType]
            Per dictionary words
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.words
        return data

    def punctuation(self):
        """
        Generate mapping for dictionaries to punctuation

        Returns
        -------
        Dict[str, str]
            Per dictionary punctuation
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.punctuation
        return data

    def clitic_set(self) -> Dict[str, Set[str]]:
        """
        Generate mapping for dictionaries to clitic sets

        Returns
        -------
        Dict[str, str]
            Per dictionary clitic sets
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.clitic_set
        return data

    def clitic_markers(self) -> Dict[str, str]:
        """
        Generate mapping for dictionaries to clitic markers

        Returns
        -------
        Dict[str, str]
            Per dictionary clitic markers
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.clitic_markers
        return data

    def compound_markers(self) -> Dict[str, str]:
        """
        Generate mapping for dictionaries to compound markers

        Returns
        -------
        Dict[str, str]
            Per dictionary compound markers
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.compound_markers
        return data

    def strip_diacritics(self) -> Dict[str, List[str]]:
        """
        Generate mapping for dictionaries to diacritics to strip

        Returns
        -------
        Dict[str, List[str]]
            Per dictionary strip diacritics
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.strip_diacritics
        return data

    def oov_codes(self) -> Dict[str, str]:
        """
        Generate mapping for dictionaries to oov symbols

        Returns
        -------
        Dict[str, str]
            Per dictionary oov symbols
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.oov_code
        return data

    def oov_ints(self) -> Dict[str, int]:
        """
        Generate mapping for dictionaries to oov ints

        Returns
        -------
        Dict[str, int]
            Per dictionary oov ints
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.oov_int
        return data

    def positions(self) -> Dict[str, List[str]]:
        """
        Generate mapping for dictionaries to positions

        Returns
        -------
        Dict[str, List[str]]
            Per dictionary positions
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.positions
        return data

    def silences(self) -> Dict[str, Set[str]]:
        """
        Generate mapping for dictionaries to silence symbols

        Returns
        -------
        Dict[str, Set[str]]
            Per dictionary silence symbols
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.silences
        return data

    def multilingual_ipa(self) -> Dict[str, bool]:
        """
        Generate mapping for dictionaries to multilingual IPA flags

        Returns
        -------
        Dict[str, bool]
            Per dictionary multilingual IPA flags
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.multilingual_ipa
        return data

    def generate_pronunciations_arguments(
        self, aligner: AlignerType
    ) -> GeneratePronunciationsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.pronunciations.generate_pronunciations_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        GeneratePronunciationsArguments
            Arguments for processing
        """
        return GeneratePronunciationsArguments(
            os.path.join(
                aligner.working_log_directory, f"generate_pronunciations.{self.name}.log"
            ),
            self.current_dictionary_names,
            self.construct_path_dictionary(aligner.data_directory, "text", "int.scp"),
            self.word_boundary_int_files(),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            aligner.current_model_path,
            self.construct_path_dictionary(aligner.working_directory, "prons", "scp"),
        )

    def alignment_improvement_arguments(
        self, aligner: AlignerType
    ) -> AlignmentImprovementArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compute_alignment_improvement_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AlignmentImprovementArguments
            Arguments for processing
        """
        return AlignmentImprovementArguments(
            os.path.join(aligner.working_log_directory, f"alignment_analysis.{self.name}.log"),
            self.current_dictionary_names,
            aligner.current_model_path,
            self.construct_path_dictionary(aligner.data_directory, "text", "int.scp"),
            self.word_boundary_int_files(),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.feature_config.frame_shift,
            self.reversed_phone_mappings(),
            self.positions(),
            self.construct_path_dictionary(
                aligner.working_directory, f"phone.{aligner.iteration}", "ctm"
            ),
        )

    def ali_to_word_ctm_arguments(self, aligner: AlignerType) -> AliToCtmArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.ali_to_ctm_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AliToCtmArguments
            Arguments for processing
        """
        return AliToCtmArguments(
            os.path.join(aligner.working_log_directory, f"get_word_ctm.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.data_directory, "text", "int.scp"),
            self.word_boundary_int_files(),
            round(self.feature_config.frame_shift / 1000, 4),
            aligner.alignment_model_path,
            self.construct_path_dictionary(aligner.working_directory, "word", "ctm"),
            True,
        )

    def ali_to_phone_ctm_arguments(self, aligner: AlignerType) -> AliToCtmArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.ali_to_ctm_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AliToCtmArguments
            Arguments for processing
        """
        return AliToCtmArguments(
            os.path.join(aligner.working_log_directory, f"get_phone_ctm.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.data_directory, "text", "int.scp"),
            self.word_boundary_int_files(),
            round(self.feature_config.frame_shift / 1000, 4),
            aligner.alignment_model_path,
            self.construct_path_dictionary(aligner.working_directory, "phone", "ctm"),
            False,
        )

    def job_utts(self) -> Dict[str, Dict[str, Utterance]]:
        """
        Generate utterances by dictionary name for the Job

        Returns
        -------
        Dict[str, Dict[str, Utterance]]
            Mapping of dictionary name to Utterance mappings
        """
        data = {}
        speakers = self.subset_speakers
        if not speakers:
            speakers = self.speakers
        for s in speakers:
            if s.dictionary.name not in data:
                data[s.dictionary.name] = {}
            data[s.dictionary.name].update(s.utterances)
        return data

    def job_files(self) -> Dict[str, File]:
        """
        Generate files for the Job

        Returns
        -------
        Dict[str, File]
            Mapping of file name to File objects
        """
        data = {}
        speakers = self.subset_speakers
        if not speakers:
            speakers = self.speakers
        for s in speakers:
            for f in s.files:
                for sf in f.speaker_ordering:
                    if sf.name == s.name:
                        sf.dictionary_data = s.dictionary_data
                data[f.name] = f
        return data

    def cleanup_word_ctm_arguments(self, aligner: AlignerType) -> CleanupWordCtmArguments:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.CleanupWordCtmProcessWorker`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        CleanupWordCtmArguments
            Arguments for processing
        """
        return CleanupWordCtmArguments(
            self.construct_path_dictionary(aligner.align_directory, "word", "ctm"),
            self.current_dictionary_names,
            self.job_utts(),
            self.dictionary_data(),
        )

    def no_cleanup_word_ctm_arguments(self, aligner: AlignerType) -> NoCleanupWordCtmArguments:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.NoCleanupWordCtmProcessWorker`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        NoCleanupWordCtmArguments
            Arguments for processing
        """
        return NoCleanupWordCtmArguments(
            self.construct_path_dictionary(aligner.align_directory, "word", "ctm"),
            self.current_dictionary_names,
            self.job_utts(),
            self.dictionary_data(),
        )

    def phone_ctm_arguments(self, aligner: AlignerType) -> PhoneCtmArguments:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.PhoneCtmProcessWorker`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        PhoneCtmArguments
            Arguments for processing
        """
        return PhoneCtmArguments(
            self.construct_path_dictionary(aligner.align_directory, "phone", "ctm"),
            self.current_dictionary_names,
            self.job_utts(),
            self.reversed_phone_mappings(),
            self.positions(),
        )

    def dictionary_data(self) -> Dict[str, DictionaryData]:
        """
        Generate dictionary data for the job

        Returns
        -------
        Dict[str, DictionaryData]
            Mapping of dictionary name to dictionary data
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = dictionary.data()
        return data

    def combine_ctm_arguments(self, aligner: AlignerType) -> CombineCtmArguments:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.CombineProcessWorker`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        CombineCtmArguments
            Arguments for processing
        """
        return CombineCtmArguments(
            self.current_dictionary_names,
            self.job_files(),
            self.dictionary_data(),
            aligner.align_config.cleanup_textgrids,
        )

    def export_textgrid_arguments(self, aligner: AlignerType) -> ExportTextGridArguments:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.multiprocessing.alignment.ExportTextGridProcessWorker`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        ExportTextGridArguments
            Arguments for processing
        """
        return ExportTextGridArguments(
            aligner.corpus.files,
            aligner.feature_config.frame_shift,
            aligner.textgrid_output,
            aligner.backup_output_directory,
        )

    def tree_stats_arguments(self, aligner: AlignerType) -> TreeStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.tree_stats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        TreeStatsArguments
            Arguments for processing
        """
        return TreeStatsArguments(
            os.path.join(aligner.working_log_directory, f"acc_tree.{self.name}.log"),
            self.current_dictionary_names,
            aligner.dictionary.silence_csl,
            aligner.previous_trainer.alignment_model_path,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(aligner.previous_trainer.align_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.working_directory, "tree", "acc"),
        )

    def convert_alignment_arguments(self, aligner: AlignerType) -> ConvertAlignmentsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.convert_alignments_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        ConvertAlignmentsArguments
            Arguments for processing
        """
        return ConvertAlignmentsArguments(
            os.path.join(aligner.working_log_directory, f"convert_alignments.{self.name}.log"),
            self.current_dictionary_names,
            aligner.current_model_path,
            aligner.tree_path,
            aligner.previous_trainer.alignment_model_path,
            self.construct_path_dictionary(
                aligner.previous_trainer.working_directory, "ali", "ark"
            ),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
        )

    def calc_fmllr_arguments(self, aligner: AlignerType) -> CalcFmllrArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.calc_fmllr_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        CalcFmllrArguments
            Arguments for processing
        """
        return CalcFmllrArguments(
            os.path.join(aligner.working_log_directory, f"calc_fmllr.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            aligner.alignment_model_path,
            aligner.current_model_path,
            self.construct_path_dictionary(aligner.data_directory, "spk2utt", "scp"),
            self.construct_path_dictionary(aligner.working_directory, "trans", "ark"),
            aligner.fmllr_options,
        )

    def acc_stats_two_feats_arguments(self, aligner: AlignerType) -> AccStatsTwoFeatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.acc_stats_two_feats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        AccStatsTwoFeatsArguments
            Arguments for processing
        """
        return AccStatsTwoFeatsArguments(
            os.path.join(aligner.working_log_directory, f"acc_stats_two_feats.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.working_directory, "two_feat_acc", "ark"),
            aligner.current_model_path,
            self.construct_feature_proc_strings(aligner),
            self.construct_feature_proc_strings(aligner, speaker_independent=True),
        )

    def lda_acc_stats_arguments(self, aligner: LdaTrainer) -> LdaAccStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.lda_acc_stats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.LdaTrainer`
            Aligner

        Returns
        -------
        LdaAccStatsArguments
            Arguments for processing
        """
        return LdaAccStatsArguments(
            os.path.join(aligner.working_log_directory, f"lda_acc_stats.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(
                aligner.previous_trainer.working_directory, "ali", "ark"
            ),
            aligner.previous_trainer.alignment_model_path,
            aligner.lda_options,
            self.construct_path_dictionary(aligner.working_directory, "lda", "acc"),
        )

    def calc_lda_mllt_arguments(self, aligner: LdaTrainer) -> CalcLdaMlltArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.calc_lda_mllt_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.LdaTrainer`
            Aligner

        Returns
        -------
        CalcLdaMlltArguments
            Arguments for processing
        """
        return CalcLdaMlltArguments(
            os.path.join(aligner.working_log_directory, f"lda_mllt.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            self.construct_path_dictionary(aligner.working_directory, "ali", "ark"),
            aligner.current_model_path,
            aligner.lda_options,
            self.construct_path_dictionary(aligner.working_directory, "lda", "macc"),
        )

    def ivector_acc_stats_arguments(
        self, trainer: IvectorExtractorTrainer
    ) -> AccIvectorStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.acc_ivector_stats_func`

        Parameters
        ----------
        trainer: :class:`~montreal_forced_aligner.trainer.IvectorExtractorTrainer`
            Aligner

        Returns
        -------
        AccIvectorStatsArguments
            Arguments for processing
        """
        return AccIvectorStatsArguments(
            os.path.join(trainer.working_log_directory, f"ivector_acc.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(trainer),
            trainer.ivector_options,
            trainer.current_ie_path,
            self.construct_path_dictionary(trainer.working_directory, "post", "ark"),
            self.construct_path_dictionary(trainer.working_directory, "ivector", "acc"),
        )

    def map_acc_stats_arguments(self, aligner: AdaptingAligner) -> MapAccStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.map_acc_stats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.aligner.AdaptingAligner`
            Aligner

        Returns
        -------
        MapAccStatsArguments
            Arguments for processing
        """
        return MapAccStatsArguments(
            os.path.join(aligner.working_log_directory, f"map_acc_stats.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            aligner.current_model_path,
            self.construct_path_dictionary(aligner.previous_aligner.align_directory, "ali", "ark"),
            self.construct_path_dictionary(aligner.working_directory, "map", "acc"),
        )

    def gmm_gselect_arguments(self, aligner: IvectorExtractorTrainer) -> GmmGselectArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.gmm_gselect_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.IvectorExtractorTrainer`
            Aligner

        Returns
        -------
        GmmGselectArguments
            Arguments for processing
        """
        return GmmGselectArguments(
            os.path.join(aligner.working_log_directory, f"gmm_gselect.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            aligner.ivector_options,
            aligner.current_dubm_path,
            self.construct_path_dictionary(aligner.working_directory, "gselect", "ark"),
        )

    def acc_global_stats_arguments(
        self, aligner: IvectorExtractorTrainer
    ) -> AccGlobalStatsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.acc_global_stats_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.IvectorExtractorTrainer`
            Aligner

        Returns
        -------
        AccGlobalStatsArguments
            Arguments for processing
        """
        return AccGlobalStatsArguments(
            os.path.join(
                aligner.working_log_directory,
                f"acc_global_stats.{aligner.iteration}.{self.name}.log",
            ),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            aligner.ivector_options,
            self.construct_path_dictionary(aligner.working_directory, "gselect", "ark"),
            self.construct_path_dictionary(
                aligner.working_directory, f"global.{aligner.iteration}", "acc"
            ),
            aligner.current_dubm_path,
        )

    def gauss_to_post_arguments(self, aligner: IvectorExtractorTrainer) -> GaussToPostArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.ivector.gauss_to_post_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.IvectorExtractorTrainer`
            Aligner

        Returns
        -------
        GaussToPostArguments
            Arguments for processing
        """
        return GaussToPostArguments(
            os.path.join(aligner.working_log_directory, f"gauss_to_post.{self.name}.log"),
            self.current_dictionary_names,
            self.construct_feature_proc_strings(aligner),
            aligner.ivector_options,
            self.construct_path_dictionary(aligner.working_directory, "post", "ark"),
            aligner.current_dubm_path,
        )

    def compile_train_graph_arguments(self, aligner: AlignerType) -> CompileTrainGraphsArguments:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.multiprocessing.alignment.compile_train_graphs_func`

        Parameters
        ----------
        aligner: :class:`~montreal_forced_aligner.trainer.BaseTrainer` or :class:`~montreal_forced_aligner.aligner.BaseAligner`
            Aligner

        Returns
        -------
        CompileTrainGraphsArguments
            Arguments for processing
        """
        dictionary_paths = aligner.dictionary.output_paths
        disambig_paths = {
            k: os.path.join(v, "phones", "disambiguation_symbols.int")
            for k, v in dictionary_paths.items()
        }
        lexicon_fst_paths = {k: os.path.join(v, "L.fst") for k, v in dictionary_paths.items()}
        model_path = aligner.current_model_path
        if not os.path.exists(model_path):
            model_path = aligner.alignment_model_path
        return CompileTrainGraphsArguments(
            os.path.join(aligner.working_log_directory, f"compile_train_graphs.{self.name}.log"),
            self.current_dictionary_names,
            os.path.join(aligner.working_directory, "tree"),
            model_path,
            self.construct_path_dictionary(aligner.data_directory, "text", "int.scp"),
            disambig_paths,
            lexicon_fst_paths,
            self.construct_path_dictionary(aligner.working_directory, "fsts", "scp"),
        )

    def utt2fst_scp_data(
        self, corpus: Corpus, num_frequent_words: int = 10
    ) -> Dict[str, List[Tuple[str, str]]]:
        """
        Generate Kaldi style utt2fst scp data

        Parameters
        ----------
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to generate data for
        num_frequent_words: int
            Number of frequent words to include in the unigram language model

        Returns
        -------
        Dict[str, List[Tuple[str, str]]]
            Utterance FSTs per dictionary name
        """
        data = {}
        most_frequent = {}
        for dict_name, utterances in self.job_utts().items():
            data[dict_name] = []
            for u_name, utterance in utterances.items():
                new_text = []
                dictionary = utterance.speaker.dictionary
                if dictionary.name not in most_frequent:
                    word_frequencies = corpus.get_word_frequency(dictionary)
                    most_frequent[dictionary.name] = sorted(
                        word_frequencies.items(), key=lambda x: -x[1]
                    )[:num_frequent_words]

                for t in utterance.text:
                    lookup = utterance.speaker.dictionary.split_clitics(t)
                    if lookup is None:
                        continue
                    new_text.extend(x for x in lookup if x != "")
                data[dict_name].append(
                    (
                        u_name,
                        dictionary.create_utterance_fst(new_text, most_frequent[dictionary.name]),
                    )
                )
        return data

    def output_utt_fsts(self, corpus: Corpus, num_frequent_words: int = 10) -> None:
        """
        Write utterance FSTs

        Parameters
        ----------
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to generate FSTs for
        num_frequent_words: int
            Number of frequent words
        """
        utt2fst = self.utt2fst_scp_data(corpus, num_frequent_words)
        for dict_name, scp in utt2fst.items():
            utt2fst_scp_path = os.path.join(
                corpus.split_directory, f"utt2fst.{dict_name}.{self.name}.scp"
            )
            save_scp(scp, utt2fst_scp_path, multiline=True)

    def output_to_directory(self, split_directory: str) -> None:
        """
        Output job information to a directory

        Parameters
        ----------
        split_directory: str
            Directory to output to
        """
        wav = self.wav_scp_data()
        for dict_name, scp in wav.items():
            wav_scp_path = os.path.join(split_directory, f"wav.{dict_name}.{self.name}.scp")
            output_mapping(scp, wav_scp_path, skip_safe=True)

        spk2utt = self.spk2utt_scp_data()
        for dict_name, scp in spk2utt.items():
            spk2utt_scp_path = os.path.join(
                split_directory, f"spk2utt.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, spk2utt_scp_path)

        feats = self.feat_scp_data()
        for dict_name, scp in feats.items():
            feats_scp_path = os.path.join(split_directory, f"feats.{dict_name}.{self.name}.scp")
            output_mapping(scp, feats_scp_path)

        cmvn = self.cmvn_scp_data()
        for dict_name, scp in cmvn.items():
            cmvn_scp_path = os.path.join(split_directory, f"cmvn.{dict_name}.{self.name}.scp")
            output_mapping(scp, cmvn_scp_path)

        utt2spk = self.utt2spk_scp_data()
        for dict_name, scp in utt2spk.items():
            utt2spk_scp_path = os.path.join(
                split_directory, f"utt2spk.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, utt2spk_scp_path)

        segments = self.segments_scp_data()
        for dict_name, scp in segments.items():
            segments_scp_path = os.path.join(
                split_directory, f"segments.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, segments_scp_path)

        text_scp = self.text_scp_data()
        for dict_name, scp in text_scp.items():
            if not scp:
                continue
            text_scp_path = os.path.join(split_directory, f"text.{dict_name}.{self.name}.scp")
            output_mapping(scp, text_scp_path)

        text_int = self.text_int_scp_data()
        for dict_name, scp in text_int.items():
            if dict_name is None:
                continue
            if not scp:
                continue
            text_int_scp_path = os.path.join(
                split_directory, f"text.{dict_name}.{self.name}.int.scp"
            )
            output_mapping(scp, text_int_scp_path, skip_safe=True)



"""Class definitions for Speaker Adapted Triphone trainer"""
from __future__ import annotations

import os
import shutil
import subprocess
import time
from typing import TYPE_CHECKING, Optional

from ..exceptions import KaldiProcessingError
from ..multiprocessing import (
    acc_stats,
    align,
    calc_fmllr,
    compile_information,
    compile_train_graphs,
    compute_alignment_improvement,
    convert_alignments,
    create_align_model,
    tree_stats,
)
from ..utils import log_kaldi_errors, parse_logs, thirdparty_binary
from .triphone import TriphoneTrainer

if TYPE_CHECKING:
    from ..config import ConfigDict, FeatureConfig
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from .base import TrainerType


__all__ = ["SatTrainer"]


class SatTrainer(TriphoneTrainer):
    """

    Configuration class for speaker adapted training (SAT)

    Attributes
    ----------
    fmllr_update_type : str
        Type of fMLLR estimation, defaults to ``'full'``
    fmllr_iterations : list
        List of iterations to perform fMLLR estimation
    silence_weight : float
        Weight on silence in fMLLR estimation
    """

    def __init__(self, default_feature_config: FeatureConfig):
        super(SatTrainer, self).__init__(default_feature_config)
        self.fmllr_update_type = "full"
        self.fmllr_iterations = []
        max_fmllr_iter = int(self.num_iterations / 2) - 1
        for i in range(1, max_fmllr_iter):
            if i < max_fmllr_iter / 2 and i % 2 == 0:
                self.fmllr_iterations.append(i)
        self.fmllr_iterations.append(max_fmllr_iter)
        self.silence_weight = 0.0
        self.feature_config.fmllr = True
        self.initial_fmllr = True
        self.ensure_train = True

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations, initial gaussians, and fMLLR iteraction based on configuration"""
        super(SatTrainer, self).compute_calculated_properties()
        self.fmllr_iterations = []
        max_fmllr_iter = int(self.num_iterations / 2) - 1
        for i in range(1, max_fmllr_iter):
            if i < max_fmllr_iter / 2 and i % 2 == 0:
                self.fmllr_iterations.append(i)
        self.fmllr_iterations.append(max_fmllr_iter)

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "sat"

    @property
    def fmllr_options(self) -> ConfigDict:
        """Options for calculating fMLLR transforms"""
        return {
            "fmllr_update_type": self.fmllr_update_type,
            "debug": self.debug,
            "initial": self.initial_fmllr,
            "silence_csl": self.dictionary.silence_csl,
        }

    @property
    def working_directory(self) -> str:
        """Current working directory"""
        if self.ensure_train:
            return self.train_directory
        return super().working_directory

    @property
    def working_log_directory(self) -> str:
        """Current log directory"""
        if self.ensure_train:
            return self.log_directory
        return super().working_log_directory

    def finalize_training(self) -> None:
        """
        Finalize training and create a speaker independent model for initial alignment

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        try:
            super().finalize_training()
            create_align_model(self)
            self.ensure_train = False
            shutil.copyfile(
                os.path.join(self.train_directory, "final.alimdl"),
                os.path.join(self.align_directory, "final.alimdl"),
            )
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def training_iteration(self) -> None:
        """
        Run a single training iteration
        """
        if os.path.exists(self.next_model_path):
            self.iteration += 1
            return
        if self.iteration in self.realignment_iterations:
            align(self)
            if self.debug:
                compute_alignment_improvement(self)
        if self.iteration in self.fmllr_iterations:
            calc_fmllr(self)

        acc_stats(self)
        parse_logs(self.log_directory)
        if self.iteration < self.final_gaussian_iteration:
            self.increment_gaussians()
        self.iteration += 1

    def align(self, subset: Optional[int] = None) -> None:
        """
        Align a given subset of the corpus

        Parameters
        ----------
        subset: int, optional
            Number of utterances to select for the aligned subset

        Raises
        ------
        KaldiProcessingError
            If there were any errors in running Kaldi binaries
        """
        if not os.path.exists(self.align_directory):
            self.finalize_training()
        dirty_path = os.path.join(self.align_directory, "dirty")
        if os.path.exists(dirty_path):  # if there was an error, let's redo from scratch
            shutil.rmtree(self.align_directory)
        done_path = os.path.join(self.align_directory, "done")
        if not os.path.exists(done_path):
            message = f"Generating alignments using {self.identifier} models"
            if subset:
                message += f" using {subset} utterances..."
            else:
                message += " for the whole corpus..."
            self.logger.info(message)
            begin = time.time()
            if subset is None:
                self.data_directory = self.corpus.split_directory
            else:
                self.data_directory = self.corpus.subset_directory(subset)
            try:
                self.speaker_independent = True
                self.initial_fmllr = True
                compile_train_graphs(self)
                align(self)

                unaligned, average_log_like = compile_information(self)
                self.logger.debug(
                    f"Before SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
                )

                if self.speaker_independent:
                    calc_fmllr(self)
                    self.speaker_independent = False
                    self.initial_fmllr = False
                    align(self)
                self.save(os.path.join(self.align_directory, "acoustic_model.zip"))

                unaligned, average_log_like = compile_information(self)
                self.logger.debug(
                    f"Following SAT, average per frame likelihood (this might not actually mean anything): {average_log_like}"
                )
            except Exception as e:
                with open(dirty_path, "w"):
                    pass
                if isinstance(e, KaldiProcessingError):
                    log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
                raise
            with open(done_path, "w"):
                pass
            self.logger.debug(f"Alignment took {time.time() - begin} seconds")
        else:
            self.logger.info(f"Alignments using {self.identifier} models already done")

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[TrainerType],
    ) -> None:
        """
        Initialize speaker-adapted triphone training

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: :class:`~montreal_forced_aligner.trainers.base.BaseTrainer`, optional
            Previous trainer to initialize from
        """
        self.feature_config.fmllr = False
        self._setup_for_init(identifier, temporary_directory, corpus, dictionary, previous_trainer)
        done_path = os.path.join(self.train_directory, "done")
        dirty_path = os.path.join(self.train_directory, "dirty")
        self.feature_config.fmllr = True
        if os.path.exists(done_path):
            self.logger.info(f"{self.identifier} training already done, skipping initialization.")
            return
        if os.path.exists(os.path.join(self.train_directory, "1.mdl")):
            return
        begin = time.time()
        self.logger.info("Initializing speaker-adapted triphone training...")
        align_directory = previous_trainer.align_directory
        try:
            if os.path.exists(os.path.join(align_directory, "lda.mat")):
                shutil.copyfile(
                    os.path.join(align_directory, "lda.mat"),
                    os.path.join(self.train_directory, "lda.mat"),
                )
            tree_stats(self)
            log_path = os.path.join(self.log_directory, "questions.log")
            tree_path = os.path.join(self.train_directory, "tree")
            treeacc_path = os.path.join(self.train_directory, "treeacc")
            sets_int_path = os.path.join(self.dictionary.phones_dir, "sets.int")
            roots_int_path = os.path.join(self.dictionary.phones_dir, "roots.int")
            extra_question_int_path = os.path.join(
                self.dictionary.phones_dir, "extra_questions.int"
            )
            topo_path = os.path.join(self.dictionary.output_directory, "topo")
            questions_path = os.path.join(self.train_directory, "questions.int")
            questions_qst_path = os.path.join(self.train_directory, "questions.qst")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("cluster-phones"),
                        treeacc_path,
                        sets_int_path,
                        questions_path,
                    ],
                    stderr=log_file,
                )

            with open(extra_question_int_path, "r") as in_file, open(
                questions_path, "a"
            ) as out_file:
                for line in in_file:
                    out_file.write(line)

            log_path = os.path.join(self.log_directory, "compile_questions.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("compile-questions"),
                        topo_path,
                        questions_path,
                        questions_qst_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "build_tree.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("build-tree"),
                        "--verbose=1",
                        f"--max-leaves={self.initial_gaussians}",
                        f"--cluster-thresh={self.cluster_threshold}",
                        treeacc_path,
                        roots_int_path,
                        questions_qst_path,
                        topo_path,
                        tree_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "init_model.log")
            occs_path = os.path.join(self.train_directory, "0.occs")
            mdl_path = os.path.join(self.train_directory, "0.mdl")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("gmm-init-model"),
                        f"--write-occs={occs_path}",
                        tree_path,
                        treeacc_path,
                        topo_path,
                        mdl_path,
                    ],
                    stderr=log_file,
                )

            log_path = os.path.join(self.log_directory, "mixup.log")
            with open(log_path, "w") as log_file:
                subprocess.call(
                    [
                        thirdparty_binary("gmm-mixup"),
                        f"--mix-up={self.initial_gaussians}",
                        mdl_path,
                        occs_path,
                        mdl_path,
                    ],
                    stderr=log_file,
                )
            os.remove(treeacc_path)

            compile_train_graphs(self)

            convert_alignments(self)

            if os.path.exists(os.path.join(align_directory, "trans.0.ark")):
                for j in self.corpus.jobs:
                    for path in j.construct_path_dictionary(
                        align_directory, "trans", "ark"
                    ).values():
                        shutil.copy(path, path.replace(align_directory, self.train_directory))
            else:

                calc_fmllr(self)
            self.initial_fmllr = False
            self.iteration = 1
            os.rename(occs_path, os.path.join(self.train_directory, "1.occs"))
            os.rename(mdl_path, os.path.join(self.train_directory, "1.mdl"))
            parse_logs(self.log_directory)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        self.logger.info("Initialization complete!")
        self.logger.debug(f"Initialization took {time.time() - begin} seconds")



"""Class definitions for configuring commands"""
from __future__ import annotations

import os
from typing import Any

import yaml

__all__ = ["CommandConfig", "load_command_configuration"]


class CommandConfig(object):
    """
    Configuration for running commands

    """

    def __init__(self, data: dict):
        self.data = data

    def __getitem__(self, item: str) -> Any:
        """Get key"""
        if item not in self.data:
            return None
        return self.data[item]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set key"""
        self.data[key] = value

    def update(self, new_data: dict) -> None:
        """Update configuration"""
        self.data.update(new_data)

    def save(self, conf_path: str) -> None:
        """Export to path"""
        with open(conf_path, "w") as f:
            yaml.dump(self.data, f)


def load_command_configuration(conf_path: str, default: dict) -> CommandConfig:
    """
    Load a previous run of MFA in a temporary directory

    Parameters
    ----------
    conf_path: str
        Path to saved configuration
    default: dict
        Extra parameters to set on load

    Returns
    -------
    :class:`~montreal_forced_aligner.config.command_config.CommandConfig`
        Command configuration
    """
    if os.path.exists(conf_path):
        with open(conf_path, "r") as f:
            conf = yaml.load(f, Loader=yaml.SafeLoader)
        config = CommandConfig(conf)
    else:
        config = CommandConfig(default)
    return config



"""Aligner module of MFA, contains various types of top-level aligners"""
from .adapting import AdaptingAligner  # noqa
from .pretrained import PretrainedAligner  # noqa
from .trainable import TrainableAligner  # noqa



"""Class definitions for trainable aligners"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .base import BaseAligner

if TYPE_CHECKING:
    from logging import Logger

    from ..aligner.pretrained import PretrainedAligner
    from ..config import AlignConfig, TrainingConfig
    from ..corpus import Corpus
    from ..dictionary import Dictionary

__all__ = ["TrainableAligner"]


class TrainableAligner(BaseAligner):
    """
    Aligner that aligns and trains acoustics models on a large dataset

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus object for the dataset
    dictionary : :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Dictionary object for the pronunciation dictionary
    training_config : :class:`~montreal_forced_aligner.config.TrainingConfig`
        Configuration to train a model
    align_config : :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Configuration for alignment
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug: bool
        Flag for debug mode, default is False
    verbose: bool
        Flag for verbose mode, default is False
    logger: :class:`~logging.Logger`
        Logger to use
    pretrained_aligner: :class:`~montreal_forced_aligner.aligner.pretrained.PretrainedAligner`, optional
        Pretrained aligner to use as input to training
    """

    def __init__(
        self,
        corpus: Corpus,
        dictionary: Dictionary,
        training_config: TrainingConfig,
        align_config: AlignConfig,
        temp_directory: Optional[str] = None,
        debug: bool = False,
        verbose: bool = False,
        logger: Optional[Logger] = None,
        pretrained_aligner: Optional[PretrainedAligner] = None,
    ):
        self.training_config = training_config
        self.pretrained_aligner = pretrained_aligner
        if self.pretrained_aligner is not None:
            acoustic_model = pretrained_aligner.acoustic_model
        else:
            acoustic_model = None
        super(TrainableAligner, self).__init__(
            corpus,
            dictionary,
            align_config,
            temp_directory,
            debug,
            verbose,
            logger,
            acoustic_model=acoustic_model,
        )
        for trainer in self.training_config.training_configs:
            trainer.logger = self.logger

    def save(self, path: str, root_directory: Optional[str] = None) -> None:
        """
        Output an acoustic model and dictionary to the specified path

        Parameters
        ----------
        path : str
            Path to save acoustic model and dictionary
        root_directory : str or None
            Path for root directory of temporary files
        """
        self.training_config.values()[-1].save(path, root_directory)
        self.logger.info("Saved model to {}".format(path))

    @property
    def meta(self) -> dict:
        """Acoustic model parameters"""
        from ..utils import get_mfa_version

        data = {
            "phones": sorted(self.dictionary.nonsil_phones),
            "version": get_mfa_version(),
            "architecture": self.training_config.values()[-1].architecture,
            "phone_type": self.training_config.values()[-1].phone_type,
            "features": self.align_config.feature_config.params(),
        }
        return data

    def train(self, generate_final_alignments: bool = True) -> None:
        """
        Run through the training configurations to produce a final acoustic model

        Parameters
        ----------
        generate_final_alignments: bool
            Flag for whether final alignments should be generated at the end of training, defaults to True
        """
        previous = self.pretrained_aligner
        for identifier, trainer in self.training_config.items():
            trainer.debug = self.debug
            trainer.logger = self.logger
            if previous is not None:
                previous.align(trainer.subset)
            trainer.init_training(
                identifier, self.temp_directory, self.corpus, self.dictionary, previous
            )
            trainer.train()
            previous = trainer
        if generate_final_alignments:
            previous.align(None)

    @property
    def align_directory(self) -> str:
        """Align directory"""
        return self.training_config.values()[-1].align_directory



"""Multiprocessing functions and classes for Montreal Forced Aligner"""
from .alignment import acc_stats  # noqa
from .alignment import align  # noqa
from .alignment import calc_fmllr  # noqa
from .alignment import calc_lda_mllt  # noqa
from .alignment import compile_information  # noqa
from .alignment import compile_train_graphs  # noqa
from .alignment import compute_alignment_improvement  # noqa
from .alignment import convert_ali_to_textgrids  # noqa
from .alignment import convert_alignments  # noqa
from .alignment import create_align_model  # noqa
from .alignment import lda_acc_stats  # noqa
from .alignment import mono_align_equal  # noqa
from .alignment import train_map  # noqa
from .alignment import tree_stats  # noqa; noqa
from .helper import Counter, Stopped, run_mp, run_non_mp  # noqa
from .ivector import acc_global_stats  # noqa
from .ivector import acc_ivector_stats  # noqa
from .ivector import extract_ivectors  # noqa
from .ivector import gauss_to_post  # noqa
from .ivector import gmm_gselect  # noqa
from .ivector import segment_vad  # noqa
from .pronunciations import generate_pronunciations  # noqa
from .transcription import transcribe, transcribe_fmllr  # noqa



"""Class definitions for configuring transcription"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING

import yaml

from .base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_DIGRAPHS,
    DEFAULT_PUNCTUATION,
    DEFAULT_STRIP_DIACRITICS,
    BaseConfig,
    ConfigError,
)
from .feature_config import FeatureConfig

if TYPE_CHECKING:
    from ..config import ConfigDict

__all__ = ["TranscribeConfig", "transcribe_yaml_to_config", "load_basic_transcribe"]


class TranscribeConfig(BaseConfig):
    """
    Class for storing metadata to configure transcription

    Parameters
    ----------
    feature_config: :class:`~montreal_forced_aligner.config.feature.FeatureConfig`
        Feature configuration to use in transcription

    Attributes
    ----------
    transition_scale: float
    acoustic_scale: float
    self_loop_scale: float
    silence_weight: float
    beam: int
    max_active: int
    fmllr: bool
    fmllr_update_type: str
    lattice_beam: int
    first_beam: int, optional
    """

    def __init__(self, feature_config: FeatureConfig):
        self.transition_scale = 1.0
        self.acoustic_scale = 0.083333
        self.self_loop_scale = 0.1
        self.feature_config = feature_config
        self.silence_weight = 0.01
        self.beam = 10
        self.max_active = 7000
        self.fmllr = True
        self.fmllr_update_type = "full"
        self.lattice_beam = 6
        self.first_beam = None
        self.first_max_active = 2000
        self.language_model_weight = 10
        self.word_insertion_penalty = 0.5
        self.data_directory = None  # Gets set later
        self.use_mp = True
        self.use_fmllr_mp = False
        self.multilingual_ipa = False
        self.ignore_speakers = False
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        self.strip_diacritics = DEFAULT_STRIP_DIACRITICS
        self.digraphs = DEFAULT_DIGRAPHS
        self.overwrite = False

    def params(self) -> ConfigDict:
        """Metadata parameters for the configuration"""
        return {
            "transition_scale": self.transition_scale,
            "acoustic_scale": self.acoustic_scale,
            "self_loop_scale": self.self_loop_scale,
            "silence_weight": self.silence_weight,
            "beam": self.beam,
            "max_active": self.max_active,
            "fmllr": self.fmllr,
            "fmllr_update_type": self.fmllr_update_type,
            "lattice_beam": self.lattice_beam,
            "first_beam": self.first_beam,
            "first_max_active": self.first_max_active,
            "language_model_weight": self.language_model_weight,
            "word_insertion_penalty": self.word_insertion_penalty,
            "use_mp": self.use_mp,
        }

    @property
    def decode_options(self) -> ConfigDict:
        """Options needed for decoding"""
        return {
            "fmllr": self.fmllr,
            "ignore_speakers": self.ignore_speakers,
            "first_beam": self.first_beam,
            "beam": self.beam,
            "first_max_active": self.first_max_active,
            "max_active": self.max_active,
            "lattice_beam": self.lattice_beam,
            "acoustic_scale": self.acoustic_scale,
        }

    @property
    def score_options(self) -> ConfigDict:
        """Options needed for scoring lattices"""
        return {
            "language_model_weight": self.language_model_weight,
            "word_insertion_penalty": self.word_insertion_penalty,
        }

    @property
    def fmllr_options(self) -> ConfigDict:
        """Options needed for calculating fMLLR transformations"""
        return {
            "fmllr_update_type": self.fmllr_update_type,
            "acoustic_scale": self.acoustic_scale,
            "silence_weight": self.silence_weight,
            "lattice_beam": self.lattice_beam,
        }

    @property
    def lm_rescore_options(self) -> ConfigDict:
        """Options needed for rescoring the language model"""
        return {
            "acoustic_scale": self.acoustic_scale,
        }

    def update(self, data: dict) -> None:
        """Update configuration with new parameters"""
        for k, v in data.items():
            if k == "use_mp":
                self.feature_config.use_mp = v
            if not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)


def transcribe_yaml_to_config(path: str) -> TranscribeConfig:
    """
    Helper function to load transcription configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.transcribe_config.TranscribeConfig`
        Transcription configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        feature_config = FeatureConfig()
        for k, v in data.items():
            if k == "features":
                feature_config.update(v)
            else:
                global_params[k] = v
        config = TranscribeConfig(feature_config)
        config.update(global_params)
        return config


def load_basic_transcribe() -> TranscribeConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.transcribe_config.TranscribeConfig`
        Default transcription configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    config = transcribe_yaml_to_config(os.path.join(base_dir, "basic_transcribe.yaml"))
    return config



"""Class definitions for acoustic model trainers in MFA"""
from .base import BaseTrainer  # noqa
from .ivector_extractor import IvectorExtractorTrainer  # noqa
from .lda import LdaTrainer  # noqa
from .monophone import MonophoneTrainer  # noqa
from .sat import SatTrainer  # noqa
from .triphone import TriphoneTrainer  # noqa

__all__ = [
    "BaseTrainer",
    "IvectorExtractorTrainer",
    "LdaTrainer",
    "MonophoneTrainer",
    "SatTrainer",
    "TriphoneTrainer",
]



"""Command line functions for calling the root mfa command"""
from __future__ import annotations

import argparse
import atexit
import multiprocessing as mp
import sys
import time
from datetime import datetime
from typing import TYPE_CHECKING

from montreal_forced_aligner.command_line.adapt import run_adapt_model
from montreal_forced_aligner.command_line.align import run_align_corpus
from montreal_forced_aligner.command_line.anchor import run_anchor
from montreal_forced_aligner.command_line.classify_speakers import run_classify_speakers
from montreal_forced_aligner.command_line.create_segments import run_create_segments
from montreal_forced_aligner.command_line.g2p import run_g2p
from montreal_forced_aligner.command_line.model import run_model
from montreal_forced_aligner.command_line.train_acoustic_model import run_train_acoustic_model
from montreal_forced_aligner.command_line.train_dictionary import run_train_dictionary
from montreal_forced_aligner.command_line.train_g2p import run_train_g2p
from montreal_forced_aligner.command_line.train_ivector_extractor import (
    run_train_ivector_extractor,
)
from montreal_forced_aligner.command_line.train_lm import run_train_lm
from montreal_forced_aligner.command_line.transcribe import run_transcribe_corpus
from montreal_forced_aligner.command_line.validate import run_validate_corpus
from montreal_forced_aligner.config import (
    load_command_history,
    load_global_config,
    update_command_history,
    update_global_config,
)
from montreal_forced_aligner.exceptions import MFAError
from montreal_forced_aligner.models import MODEL_TYPES
from montreal_forced_aligner.utils import (
    get_available_acoustic_models,
    get_available_dictionaries,
    get_available_g2p_models,
    get_available_ivector_extractors,
    get_available_language_models,
)

if TYPE_CHECKING:
    from argparse import ArgumentParser


BEGIN = time.time()
BEGIN_DATE = datetime.now()


__all__ = ["ExitHooks", "history_save_handler", "create_parser", "main"]


class ExitHooks(object):
    """
    Class for capturing exit information for MFA commands
    """

    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        """Hook for capturing information about exit code and exceptions"""
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        """Actual exit for the program"""
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, *args):
        """Handle and save exceptions"""
        self.exception = exc


def history_save_handler() -> None:
    """
    Handler for saving history on exit.  In addition to the command run, also saves exit code, whether
    an exception was encountered, when the command was executed, and how long it took to run
    """
    from montreal_forced_aligner.utils import get_mfa_version

    history_data = {
        "command": " ".join(sys.argv),
        "execution_time": time.time() - BEGIN,
        "date": BEGIN_DATE,
        "version": get_mfa_version(),
    }

    if hooks.exit_code is not None:
        history_data["exit_code"] = hooks.exit_code
        history_data["exception"] = ""
    elif hooks.exception is not None:
        history_data["exit_code"] = 1
        history_data["exception"] = str(hooks.exception)
    else:
        history_data["exception"] = ""
        history_data["exit_code"] = 0
    update_command_history(history_data)
    if hooks.exception:
        raise hooks.exception


acoustic_models = get_available_acoustic_models()
ivector_extractors = get_available_ivector_extractors()
language_models = get_available_language_models()
g2p_models = get_available_g2p_models()
dictionaries = get_available_dictionaries()


def create_parser() -> ArgumentParser:
    """
    Constructs the MFA argument parser

    Returns
    -------
    ArgumentParser
        MFA argument parser
    """
    GLOBAL_CONFIG = load_global_config()

    def add_global_options(subparser: argparse.ArgumentParser, textgrid_output: bool = False):
        """
        Add a set of global options to a subparser

        Parameters
        ----------
        subparser: argparse.ArgumentParser
            Subparser to augment
        textgrid_output: bool
            Flag for whether the subparser is used for a command that generates TextGrids
        """
        subparser.add_argument(
            "-t",
            "--temp_directory",
            type=str,
            default=GLOBAL_CONFIG["temp_directory"],
            help=f"Temporary directory root to store MFA created files, default is {GLOBAL_CONFIG['temp_directory']}",
        )
        subparser.add_argument(
            "--disable_mp",
            help=f"Disable any multiprocessing during alignment (not recommended), default is {not GLOBAL_CONFIG['use_mp']}",
            action="store_true",
            default=not GLOBAL_CONFIG["use_mp"],
        )
        subparser.add_argument(
            "-j",
            "--num_jobs",
            type=int,
            default=GLOBAL_CONFIG["num_jobs"],
            help=f"Number of data splits (and cores to use if multiprocessing is enabled), defaults "
            f"is {GLOBAL_CONFIG['num_jobs']}",
        )
        subparser.add_argument(
            "-v",
            "--verbose",
            help=f"Output debug messages, default is {GLOBAL_CONFIG['verbose']}",
            action="store_true",
            default=GLOBAL_CONFIG["verbose"],
        )
        subparser.add_argument(
            "--clean",
            help=f"Remove files from previous runs, default is {GLOBAL_CONFIG['clean']}",
            action="store_true",
            default=GLOBAL_CONFIG["clean"],
        )
        subparser.add_argument(
            "--overwrite",
            help=f"Overwrite output files when they exist, default is {GLOBAL_CONFIG['overwrite']}",
            action="store_true",
            default=GLOBAL_CONFIG["overwrite"],
        )
        subparser.add_argument(
            "--debug",
            help=f"Run extra steps for debugging issues, default is {GLOBAL_CONFIG['debug']}",
            action="store_true",
            default=GLOBAL_CONFIG["debug"],
        )
        if textgrid_output:
            subparser.add_argument(
                "--disable_textgrid_cleanup",
                help=f"Disable extra clean up steps on TextGrid output, default is {not GLOBAL_CONFIG['cleanup_textgrids']}",
                action="store_true",
                default=not GLOBAL_CONFIG["cleanup_textgrids"],
            )

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    _ = subparsers.add_parser("version")

    align_parser = subparsers.add_parser("align")
    align_parser.add_argument("corpus_directory", help="Full path to the directory to align")
    align_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use"
    )
    align_parser.add_argument(
        "acoustic_model_path",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(acoustic_models)})",
    )
    align_parser.add_argument(
        "output_directory",
        help="Full path to output directory, will be created if it doesn't exist",
    )
    align_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for alignment"
    )
    align_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of file names to use for determining speaker, "
        "default is to use directory names",
    )
    align_parser.add_argument(
        "-a",
        "--audio_directory",
        type=str,
        default="",
        help="Audio directory root to use for finding audio files",
    )
    add_global_options(align_parser, textgrid_output=True)

    adapt_parser = subparsers.add_parser("adapt")
    adapt_parser.add_argument("corpus_directory", help="Full path to the directory to align")
    adapt_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use"
    )
    adapt_parser.add_argument(
        "acoustic_model_path",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(acoustic_models)})",
    )
    adapt_parser.add_argument(
        "output_paths",
        nargs="+",
        help="Path to directory for aligned TextGrids, zip path to export acoustic model, or both",
    )
    adapt_parser.add_argument(
        "-o",
        "--output_model_path",
        type=str,
        default="",
        help="Full path to save adapted acoustic model",
    )
    adapt_parser.add_argument(
        "--full_train",
        action="store_true",
        help="Specify whether to do a round of speaker-adapted training rather than the default "
        "remapping approach to adaptation",
    )
    adapt_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for alignment"
    )
    adapt_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of file names to use for determining speaker, "
        "default is to use directory names",
    )
    adapt_parser.add_argument(
        "-a",
        "--audio_directory",
        type=str,
        default="",
        help="Audio directory root to use for finding audio files",
    )
    add_global_options(adapt_parser, textgrid_output=True)

    train_parser = subparsers.add_parser("train")
    train_parser.add_argument(
        "corpus_directory", help="Full path to the source directory to align"
    )
    train_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use", default=""
    )
    train_parser.add_argument(
        "output_paths",
        nargs="+",
        help="Path to directory for aligned TextGrids, zip path to export acoustic model, or both",
    )
    train_parser.add_argument(
        "--config_path",
        type=str,
        default="",
        help="Path to config file to use for training and alignment",
    )
    train_parser.add_argument(
        "-o",
        "--output_model_path",
        type=str,
        default="",
        help="Full path to save resulting acoustic model",
    )
    train_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of filenames to use for determining speaker, "
        "default is to use directory names",
    )
    train_parser.add_argument(
        "-a",
        "--audio_directory",
        type=str,
        default="",
        help="Audio directory root to use for finding audio files",
    )
    add_global_options(train_parser, textgrid_output=True)

    validate_parser = subparsers.add_parser("validate")
    validate_parser.add_argument(
        "corpus_directory", help="Full path to the source directory to align"
    )
    validate_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use", default=""
    )
    validate_parser.add_argument(
        "acoustic_model_path",
        nargs="?",
        default="",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(acoustic_models)})",
    )
    validate_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of file names to use for determining speaker, "
        "default is to use directory names",
    )
    validate_parser.add_argument(
        "--test_transcriptions", help="Test accuracy of transcriptions", action="store_true"
    )
    validate_parser.add_argument(
        "--ignore_acoustics",
        help="Skip acoustic feature generation and associated validation",
        action="store_true",
    )
    add_global_options(validate_parser)

    g2p_model_help_message = f"""Full path to the archive containing pre-trained model or language ({', '.join(g2p_models)})
    If not specified, then orthographic transcription is split into pronunciations."""
    g2p_parser = subparsers.add_parser("g2p")
    g2p_parser.add_argument("g2p_model_path", help=g2p_model_help_message, nargs="?")

    g2p_parser.add_argument(
        "input_path",
        help="Corpus to base word list on or a text file of words to generate pronunciations",
    )
    g2p_parser.add_argument("output_path", help="Path to save output dictionary")
    g2p_parser.add_argument(
        "--include_bracketed",
        help="Included words enclosed by brackets, job_name.e. [...], (...), <...>",
        action="store_true",
    )
    g2p_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for G2P"
    )
    add_global_options(g2p_parser)

    train_g2p_parser = subparsers.add_parser("train_g2p")
    train_g2p_parser.add_argument("dictionary_path", help="Location of existing dictionary")

    train_g2p_parser.add_argument("output_model_path", help="Desired location of generated model")
    train_g2p_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for G2P"
    )
    train_g2p_parser.add_argument(
        "--validate",
        action="store_true",
        help="Perform an analysis of accuracy training on "
        "most of the data and validating on an unseen subset",
    )
    add_global_options(train_g2p_parser)

    model_parser = subparsers.add_parser("model")

    model_subparsers = model_parser.add_subparsers(dest="action")
    model_subparsers.required = True
    model_download_parser = model_subparsers.add_parser("download")
    model_download_parser.add_argument(
        "model_type", help=f"Type of model to download, options: {', '.join(MODEL_TYPES)}"
    )
    model_download_parser.add_argument(
        "name",
        help="Name of language code to download, if not specified, "
        "will list all available languages",
        nargs="?",
    )

    model_list_parser = model_subparsers.add_parser("list")
    model_list_parser.add_argument(
        "model_type", nargs="?", help=f"Type of model to list, options: {', '.join(MODEL_TYPES)}"
    )

    model_inspect_parser = model_subparsers.add_parser("inspect")
    model_inspect_parser.add_argument(
        "model_type",
        nargs="?",
        help=f"Type of model to download, options: {', '.join(MODEL_TYPES)}",
    )
    model_inspect_parser.add_argument(
        "name", help="Name of pretrained model or path to MFA model to inspect"
    )

    model_save_parser = model_subparsers.add_parser("save")
    model_save_parser.add_argument("model_type", help="Type of MFA model")
    model_save_parser.add_argument(
        "path", help="Path to MFA model to save for invoking with just its name"
    )
    model_save_parser.add_argument(
        "--name",
        help="Name to use as reference (defaults to the name of the zip file",
        type=str,
        default="",
    )
    model_save_parser.add_argument(
        "--overwrite",
        help="Flag to overwrite existing pretrained models with the same name (and model type)",
        action="store_true",
    )

    train_lm_parser = subparsers.add_parser("train_lm")
    train_lm_parser.add_argument(
        "source_path",
        help="Full path to the source directory to train from, alternatively "
        "an ARPA format language model to convert for MFA use",
    )
    train_lm_parser.add_argument(
        "output_model_path", type=str, help="Full path to save resulting language model"
    )
    train_lm_parser.add_argument(
        "-m",
        "--model_path",
        type=str,
        help="Full path to existing language model to merge probabilities",
    )
    train_lm_parser.add_argument(
        "-w",
        "--model_weight",
        type=float,
        default=1.0,
        help="Weight factor for supplemental language model, defaults to 1.0",
    )
    train_lm_parser.add_argument(
        "--dictionary_path", help="Full path to the pronunciation dictionary to use", default=""
    )
    train_lm_parser.add_argument(
        "--config_path",
        type=str,
        default="",
        help="Path to config file to use for training and alignment",
    )
    add_global_options(train_lm_parser)

    train_dictionary_parser = subparsers.add_parser("train_dictionary")
    train_dictionary_parser.add_argument(
        "corpus_directory", help="Full path to the directory to align"
    )
    train_dictionary_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use"
    )
    train_dictionary_parser.add_argument(
        "acoustic_model_path",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(acoustic_models)})",
    )
    train_dictionary_parser.add_argument(
        "output_directory",
        help="Full path to output directory, will be created if it doesn't exist",
    )
    train_dictionary_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for alignment"
    )
    train_dictionary_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of file names to use for determining speaker, "
        "default is to use directory names",
    )
    add_global_options(train_dictionary_parser)

    train_ivector_parser = subparsers.add_parser("train_ivector")
    train_ivector_parser.add_argument(
        "corpus_directory",
        help="Full path to the source directory to " "train the ivector extractor",
    )
    train_ivector_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use"
    )
    train_ivector_parser.add_argument(
        "acoustic_model_path",
        type=str,
        default="",
        help="Full path to acoustic model for alignment",
    )
    train_ivector_parser.add_argument(
        "output_model_path",
        type=str,
        default="",
        help="Full path to save resulting ivector extractor",
    )
    train_ivector_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of filenames to use for determining speaker, "
        "default is to use directory names",
    )
    train_ivector_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for training"
    )
    add_global_options(train_ivector_parser)

    classify_speakers_parser = subparsers.add_parser("classify_speakers")
    classify_speakers_parser.add_argument(
        "corpus_directory",
        help="Full path to the source directory to " "run speaker classification",
    )
    classify_speakers_parser.add_argument(
        "ivector_extractor_path", type=str, default="", help="Full path to ivector extractor model"
    )
    classify_speakers_parser.add_argument(
        "output_directory",
        help="Full path to output directory, will be created if it doesn't exist",
    )

    classify_speakers_parser.add_argument(
        "-s", "--num_speakers", type=int, default=0, help="Number of speakers if known"
    )
    classify_speakers_parser.add_argument(
        "--cluster", help="Using clustering instead of classification", action="store_true"
    )
    classify_speakers_parser.add_argument(
        "--config_path",
        type=str,
        default="",
        help="Path to config file to use for ivector extraction",
    )
    add_global_options(classify_speakers_parser)

    create_segments_parser = subparsers.add_parser("create_segments")
    create_segments_parser.add_argument(
        "corpus_directory", help="Full path to the source directory to " "run VAD segmentation"
    )
    create_segments_parser.add_argument(
        "output_directory",
        help="Full path to output directory, will be created if it doesn't exist",
    )
    create_segments_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for segmentation"
    )
    add_global_options(create_segments_parser)

    transcribe_parser = subparsers.add_parser("transcribe")
    transcribe_parser.add_argument(
        "corpus_directory", help="Full path to the directory to transcribe"
    )
    transcribe_parser.add_argument(
        "dictionary_path", help="Full path to the pronunciation dictionary to use"
    )
    transcribe_parser.add_argument(
        "acoustic_model_path",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(acoustic_models)})",
    )
    transcribe_parser.add_argument(
        "language_model_path",
        help=f"Full path to the archive containing pre-trained model or language ({', '.join(language_models)})",
    )
    transcribe_parser.add_argument(
        "output_directory",
        help="Full path to output directory, will be created if it doesn't exist",
    )
    transcribe_parser.add_argument(
        "--config_path", type=str, default="", help="Path to config file to use for transcription"
    )
    transcribe_parser.add_argument(
        "-s",
        "--speaker_characters",
        type=str,
        default="0",
        help="Number of characters of file names to use for determining speaker, "
        "default is to use directory names",
    )
    transcribe_parser.add_argument(
        "-a",
        "--audio_directory",
        type=str,
        default="",
        help="Audio directory root to use for finding audio files",
    )
    transcribe_parser.add_argument(
        "-e",
        "--evaluate",
        help="Evaluate the transcription " "against golden texts",
        action="store_true",
    )
    add_global_options(transcribe_parser)

    config_parser = subparsers.add_parser(
        "configure",
        help="The configure command is used to set global defaults for MFA so "
        "you don't have to set them every time you call an MFA command.",
    )
    config_parser.add_argument(
        "-t",
        "--temp_directory",
        type=str,
        default="",
        help=f"Set the default temporary directory, default is {GLOBAL_CONFIG['temp_directory']}",
    )
    config_parser.add_argument(
        "-j",
        "--num_jobs",
        type=int,
        help=f"Set the number of processes to use by default, defaults to {GLOBAL_CONFIG['num_jobs']}",
    )
    config_parser.add_argument(
        "--always_clean",
        help="Always remove files from previous runs by default",
        action="store_true",
    )
    config_parser.add_argument(
        "--never_clean",
        help="Don't remove files from previous runs by default",
        action="store_true",
    )
    config_parser.add_argument(
        "--always_verbose", help="Default to verbose output", action="store_true"
    )
    config_parser.add_argument(
        "--never_verbose", help="Default to non-verbose output", action="store_true"
    )
    config_parser.add_argument(
        "--always_debug", help="Default to running debugging steps", action="store_true"
    )
    config_parser.add_argument(
        "--never_debug", help="Default to not running debugging steps", action="store_true"
    )
    config_parser.add_argument(
        "--always_overwrite", help="Always overwrite output files", action="store_true"
    )
    config_parser.add_argument(
        "--never_overwrite",
        help="Never overwrite output files (if file already exists, "
        "the output will be saved in the temp directory)",
        action="store_true",
    )
    config_parser.add_argument(
        "--disable_mp",
        help="Disable all multiprocessing (not recommended as it will usually "
        "increase processing times)",
        action="store_true",
    )
    config_parser.add_argument(
        "--enable_mp",
        help="Enable multiprocessing (recommended and enabled by default)",
        action="store_true",
    )
    config_parser.add_argument(
        "--disable_textgrid_cleanup",
        help="Disable postprocessing of TextGrids that cleans up "
        "silences and recombines compound words and clitics",
        action="store_true",
    )
    config_parser.add_argument(
        "--enable_textgrid_cleanup",
        help="Enable postprocessing of TextGrids that cleans up "
        "silences and recombines compound words and clitics",
        action="store_true",
    )
    config_parser.add_argument(
        "--disable_terminal_colors", help="Turn off colored text in output", action="store_true"
    )
    config_parser.add_argument(
        "--enable_terminal_colors", help="Turn on colored text in output", action="store_true"
    )
    config_parser.add_argument(
        "--terminal_width",
        help=f"Set width of terminal output, "
        f"currently set to {GLOBAL_CONFIG['terminal_width']}",
        default=GLOBAL_CONFIG["terminal_width"],
        type=int,
    )
    config_parser.add_argument(
        "--blas_num_threads",
        help=f"Number of threads to use for BLAS libraries, 1 is recommended "
        f"due to how much MFA relies on multiprocessing. "
        f"Currently set to {GLOBAL_CONFIG['blas_num_threads']}",
        default=GLOBAL_CONFIG["blas_num_threads"],
        type=int,
    )

    history_parser = subparsers.add_parser("history")

    history_parser.add_argument("depth", help="Number of commands to list", nargs="?", default=10)
    history_parser.add_argument(
        "--verbose", help="Flag for whether to output additional information", action="store_true"
    )

    _ = subparsers.add_parser("annotator")
    _ = subparsers.add_parser("anchor")

    return parser


parser = create_parser()


def main() -> None:
    """
    Main function for the MFA command line interface
    """
    parser = create_parser()
    mp.freeze_support()
    args, unknown = parser.parse_known_args()
    for short in ["-c", "-d"]:
        if short in unknown:
            print(
                f"Due to the number of options that `{short}` could refer to, it is not accepted. "
                "Please specify the full argument"
            )
            sys.exit(1)
    try:
        if args.subcommand in ["g2p", "train_g2p"]:
            try:
                import pynini  # noqa
            except ImportError:
                print(
                    "There was an issue importing Pynini, please ensure that it is installed. If you are on Windows, "
                    "please use the Windows Subsystem for Linux to use g2p functionality."
                )
                sys.exit(1)
        if args.subcommand == "align":
            run_align_corpus(args, unknown)
        elif args.subcommand == "adapt":
            run_adapt_model(args, unknown)
        elif args.subcommand == "train":
            run_train_acoustic_model(args, unknown)
        elif args.subcommand == "g2p":
            run_g2p(args, unknown)
        elif args.subcommand == "train_g2p":
            run_train_g2p(args, unknown)
        elif args.subcommand == "validate":
            run_validate_corpus(args, unknown)
        elif args.subcommand == "model":
            run_model(args)
        elif args.subcommand == "train_lm":
            run_train_lm(args, unknown)
        elif args.subcommand == "train_dictionary":
            run_train_dictionary(args, unknown)
        elif args.subcommand == "train_ivector":
            run_train_ivector_extractor(args, unknown)
        elif args.subcommand == "classify_speakers":
            run_classify_speakers(args, unknown)
        elif args.subcommand in ["annotator", "anchor"]:
            run_anchor()
        elif args.subcommand == "transcribe":
            run_transcribe_corpus(args, unknown)
        elif args.subcommand == "create_segments":
            run_create_segments(args, unknown)
        elif args.subcommand == "configure":
            update_global_config(args)
            global GLOBAL_CONFIG
            GLOBAL_CONFIG = load_global_config()
        elif args.subcommand == "history":
            depth = args.depth
            history = load_command_history()[-depth:]
            if args.verbose:
                print("command\tDate\tExecution time\tVersion\tExit code\tException")
                for h in history:
                    execution_time = time.strftime("%H:%M:%S", time.gmtime(h["execution_time"]))
                    d = h["date"].isoformat()
                    print(
                        f"{h['command']}\t{d}\t{execution_time}\t{h['version']}\t{h['exit_code']}\t{h['exception']}"
                    )
                pass
            else:
                for h in history:
                    print(h["command"])

        elif args.subcommand == "version":
            from montreal_forced_aligner.utils import get_mfa_version

            print(get_mfa_version())
    except MFAError as e:
        if getattr(args, "debug", False):
            raise
        print(e)
        sys.exit(1)


if __name__ == "__main__":
    hooks = ExitHooks()
    hooks.hook()
    atexit.register(history_save_handler)
    from colorama import init

    init()
    main()



"""Class definitions for configuring G2P generation"""
from __future__ import annotations

import yaml

from .base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_PUNCTUATION,
    BaseConfig,
    ConfigError,
)

__all__ = ["G2PConfig", "g2p_yaml_to_config", "load_basic_g2p_config"]


class G2PConfig(BaseConfig):
    """
    Configuration class for generating pronunciations

    """

    def __init__(self):
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        self.num_pronunciations = 1
        self.use_mp = True

    def update(self, data: dict) -> None:
        """Update configuration"""
        for k, v in data.items():
            if k in ["punctuation", "clitic_markers", "compound_markers"]:
                if not v:
                    continue
                if "-" in v:
                    v = "-" + v.replace("-", "")
                if "]" in v and r"\]" not in v:
                    v = v.replace("]", r"\]")
            elif not hasattr(self, k):
                raise ConfigError("No field found for key {}".format(k))
            setattr(self, k, v)


def g2p_yaml_to_config(path: str) -> G2PConfig:
    """
    Helper function to load G2P configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.g2p_config.G2PConfig`
        G2P configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        for k, v in data.items():
            global_params[k] = v
        g2p_config = G2PConfig()
        g2p_config.update(global_params)
        return g2p_config


def load_basic_g2p_config() -> G2PConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.g2p_config.G2PConfig`
        Default G2P configuration
    """
    return G2PConfig()



"""Command line functions for generating pronunciations using G2P models"""
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import TEMP_DIR
from montreal_forced_aligner.config.g2p_config import g2p_yaml_to_config, load_basic_g2p_config
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.dictionary import check_bracketed
from montreal_forced_aligner.g2p.generator import PyniniDictionaryGenerator as Generator
from montreal_forced_aligner.models import G2PModel
from montreal_forced_aligner.utils import setup_logger

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["generate_dictionary", "validate_args", "run_g2p"]


def generate_dictionary(args: Namespace, unknown_args: Optional[list] = None) -> None:
    """
    Run the G2P command

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    command = "g2p"
    if not args.temp_directory:
        temp_dir = TEMP_DIR
        temp_dir = os.path.join(temp_dir, "G2P")
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    if args.clean:
        shutil.rmtree(os.path.join(temp_dir, "G2P"), ignore_errors=True)
        shutil.rmtree(os.path.join(temp_dir, "models", "G2P"), ignore_errors=True)
    if args.config_path:
        g2p_config = g2p_yaml_to_config(args.config_path)
    else:
        g2p_config = load_basic_g2p_config()
    g2p_config.use_mp = not args.disable_mp
    if unknown_args:
        g2p_config.update_from_unknown_args(unknown_args)
    if os.path.isdir(args.input_path):
        input_dir = os.path.expanduser(args.input_path)
        corpus_name = os.path.basename(args.input_path)
        if corpus_name == "":
            args.input_path = os.path.dirname(args.input_path)
            corpus_name = os.path.basename(args.input_path)
        data_directory = os.path.join(temp_dir, corpus_name)
        if getattr(args, "verbose", False):
            log_level = "debug"
        else:
            log_level = "info"
        logger = setup_logger(command, data_directory, console_level=log_level)

        corpus = Corpus(
            input_dir,
            data_directory,
            num_jobs=args.num_jobs,
            use_mp=g2p_config.use_mp,
            punctuation=g2p_config.punctuation,
            clitic_markers=g2p_config.clitic_markers,
            parse_text_only_files=True,
        )

        word_set = corpus.word_set
        if not args.include_bracketed:
            word_set = [x for x in word_set if not check_bracketed(x)]
    else:

        if getattr(args, "verbose", False):
            log_level = "debug"
        else:
            log_level = "info"
        logger = setup_logger(command, temp_dir, console_level=log_level)
        word_set = []
        with open(args.input_path, "r", encoding="utf8") as f:
            for line in f:
                word_set.extend(line.strip().split())
        if not args.include_bracketed:
            word_set = [x for x in word_set if not check_bracketed(x)]

    logger.info(
        f"Generating transcriptions for the {len(word_set)} word types found in the corpus..."
    )
    if args.g2p_model_path is not None:
        model = G2PModel(
            args.g2p_model_path, root_directory=os.path.join(temp_dir, "models", "G2P")
        )
        model.validate(word_set)
        num_jobs = args.num_jobs
        if not g2p_config.use_mp:
            num_jobs = 1
        gen = Generator(
            model,
            word_set,
            temp_directory=temp_dir,
            num_jobs=num_jobs,
            num_pronunciations=g2p_config.num_pronunciations,
            logger=logger,
        )
        gen.output(args.output_path)
        model.clean_up()
    else:
        with open(args.output_path, "w", encoding="utf8") as f:
            for word in word_set:
                pronunciation = list(word)
                f.write(f"{word} {' '.join(pronunciation)}\n")


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    if not args.g2p_model_path:
        args.g2p_model_path = None
    else:
        args.g2p_model_path = validate_model_arg(args.g2p_model_path, "g2p")


def run_g2p(args: Namespace, unknown: Optional[list] = None) -> None:
    """
    Wrapper function for running G2P

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    generate_dictionary(args, unknown)
